# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch

from vllm.triton_utils import tl, triton

AWQ_TRITON_SUPPORTED_GROUP_SIZES = [-1, 32, 64, 128]


@triton.jit
def awq_dequantize_kernel(
    qweight_ptr,  # quantized matrix
    scales_ptr,  # scales, per group
    zeros_ptr,  # zeros, per group
    group_size,  # Should always be one of the supported group sizes
    result_ptr,  # Output matrix
    num_cols,  # input num cols in qweight
    num_rows,  # input num rows in qweight
    BLOCK_SIZE_X: tl.constexpr,
    BLOCK_SIZE_Y: tl.constexpr,
):
    # Set up the pids.
    pid_x = tl.program_id(axis=0)
    pid_y = tl.program_id(axis=1)

    # Compute offsets and masks for qweight_ptr.
    offsets_y = pid_y * BLOCK_SIZE_Y + tl.arange(0, BLOCK_SIZE_Y)
    offsets_x = pid_x * BLOCK_SIZE_X + tl.arange(0, BLOCK_SIZE_X)
    offsets = num_cols * offsets_y[:, None] + offsets_x[None, :]

    masks_y = offsets_y < num_rows
    masks_x = offsets_x < num_cols

    masks = masks_y[:, None] & masks_x[None, :]

    # Compute offsets and masks for result output ptr.
    result_offsets_y = pid_y * BLOCK_SIZE_Y + tl.arange(0, BLOCK_SIZE_Y)
    result_offsets_x = pid_x * BLOCK_SIZE_X * 8 + tl.arange(0, BLOCK_SIZE_X * 8)
    result_offsets = (
        8 * num_cols * result_offsets_y[:, None] + result_offsets_x[None, :]
    )

    result_masks_y = result_offsets_y < num_rows
    result_masks_x = result_offsets_x < num_cols * 8
    result_masks = result_masks_y[:, None] & result_masks_x[None, :]

    # Load the weights.
    iweights = tl.load(qweight_ptr + offsets, masks, 0.0)
    iweights = tl.interleave(iweights, iweights)
    iweights = tl.interleave(iweights, iweights)
    iweights = tl.interleave(iweights, iweights)

    # Create reverse AWQ order as tensor: [0, 4, 1, 5, 2, 6, 3, 7]
    # that will map given indices to the correct order.
    reverse_awq_order_tensor = (
        (tl.arange(0, 2) * 4)[None, :] + tl.arange(0, 4)[:, None]
    ).reshape(8)

    # Use this to compute a set of shifts that can be used to unpack and
    # reorder the values in iweights and zeros.
    shifts = reverse_awq_order_tensor * 4
    shifts = tl.broadcast_to(shifts[None, :], (BLOCK_SIZE_Y * BLOCK_SIZE_X, 8))
    shifts = tl.reshape(shifts, (BLOCK_SIZE_Y, BLOCK_SIZE_X * 8))

    # Unpack and reorder: shift out the correct 4-bit value and mask.
    iweights = (iweights >> shifts) & 0xF

    # Compute zero offsets and masks.
    zero_offsets_y = pid_y * BLOCK_SIZE_Y // group_size + tl.arange(0, 1)
    zero_offsets_x = pid_x * BLOCK_SIZE_X + tl.arange(0, BLOCK_SIZE_X)
    zero_offsets = num_cols * zero_offsets_y[:, None] + zero_offsets_x[None, :]

    zero_masks_y = zero_offsets_y < num_rows // group_size
    zero_masks_x = zero_offsets_x < num_cols
    zero_masks = zero_masks_y[:, None] & zero_masks_x[None, :]

    # Load the zeros.
    zeros = tl.load(zeros_ptr + zero_offsets, zero_masks, 0.0)
    zeros = tl.interleave(zeros, zeros)
    zeros = tl.interleave(zeros, zeros)
    zeros = tl.interleave(zeros, zeros)
    zeros = tl.broadcast_to(zeros, (BLOCK_SIZE_Y, BLOCK_SIZE_X * 8))

    # Unpack and reorder: shift out the correct 4-bit value and mask.
    zeros = (zeros >> shifts) & 0xF

    # Compute scale offsets and masks.
    scale_offsets_y = pid_y * BLOCK_SIZE_Y // group_size + tl.arange(0, 1)
    scale_offsets_x = pid_x * BLOCK_SIZE_X * 8 + tl.arange(0, BLOCK_SIZE_X * 8)
    scale_offsets = num_cols * 8 * scale_offsets_y[:, None] + scale_offsets_x[None, :]
    scale_masks_y = scale_offsets_y < num_rows // group_size
    scale_masks_x = scale_offsets_x < num_cols * 8
    scale_masks = scale_masks_y[:, None] & scale_masks_x[None, :]

    # Load the scales.
    scales = tl.load(scales_ptr + scale_offsets, scale_masks, 0.0)
    scales = tl.broadcast_to(scales, (BLOCK_SIZE_Y, BLOCK_SIZE_X * 8))

    # Dequantize.
    iweights = (iweights - zeros) * scales
    iweights = iweights.to(result_ptr.type.element_ty)

    # Finally, store.
    tl.store(result_ptr + result_offsets, iweights, result_masks)


@triton.jit
def awq_gemm_kernel(
    a_ptr,
    b_ptr,
    c_ptr,
    zeros_ptr,
    scales_ptr,
    M,
    N,
    K,
    group_size,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    BLOCK_SIZE_K: tl.constexpr,
    SPLIT_K: tl.constexpr,
):
    pid = tl.program_id(axis=0)
    pid_z = tl.program_id(1)

    # NOTE: This doesn't work in TRITON_INTERPRET=1 mode.  Use below instead.
    # num_pid_n = (N + BLOCK_SIZE_N - 1) // BLOCK_SIZE_N
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)

    pid_m = pid // num_pid_n
    pid_n = pid % num_pid_n

    accumulator_dtype = c_ptr.type.element_ty

    # NOTE: This doesn't work in TRITON_INTERPRET=1 mode.  Use below instead.
    # accumulator = tl.arange(0, BLOCK_SIZE_N)
    # accumulator = tl.broadcast_to(accumulator[None, :],
    # (BLOCK_SIZE_M, BLOCK_SIZE_N))
    # accumulator = accumulator & 0x0
    # accumulator = accumulator.to(accumulator_dtype)
    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=accumulator_dtype)

    # Create reverse AWQ order as tensor: [0, 4, 1, 5, 2, 6, 3, 7]
    # that will map given indices to the correct order.
    reverse_awq_order_tensor = (
        (tl.arange(0, 2) * 4)[None, :] + tl.arange(0, 4)[:, None]
    ).reshape(8)

    # Create the necessary shifts to use to unpack.
    shifts = reverse_awq_order_tensor * 4
    shifts = tl.broadcast_to(shifts[None, :], (BLOCK_SIZE_K * (BLOCK_SIZE_N // 8), 8))
    shifts = tl.reshape(shifts, (BLOCK_SIZE_K, BLOCK_SIZE_N))

    # Offsets and masks.
    offsets_am = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    masks_am = offsets_am < M

    offsets_bn = pid_n * (BLOCK_SIZE_N // 8) + tl.arange(0, BLOCK_SIZE_N // 8)
    masks_bn = offsets_bn < N // 8

    offsets_zn = pid_n * (BLOCK_SIZE_N // 8) + tl.arange(0, BLOCK_SIZE_N // 8)
    masks_zn = offsets_zn < N // 8

    offsets_sn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    masks_sn = offsets_sn < N

    offsets_k = pid_z * BLOCK_SIZE_K + tl.arange(0, BLOCK_SIZE_K)
    offsets_a = K * offsets_am[:, None] + offsets_k[None, :]
    offsets_b = (N // 8) * offsets_k[:, None] + offsets_bn[None, :]

    a_ptrs = a_ptr + offsets_a
    b_ptrs = b_ptr + offsets_b

    # NOTE: Use this in TRITON_INTERPRET=1 mode instead of tl.cdiv
    # block_offset = BLOCK_SIZE_K * SPLIT_K
    # for k in range(0, (K + block_offset - 1) // (block_offset)):
    for k in range(0, tl.cdiv(K, BLOCK_SIZE_K * SPLIT_K)):
        masks_k = offsets_k < K
        masks_a = masks_am[:, None] & masks_k[None, :]
        a = tl.load(a_ptrs, mask=masks_a, other=0.0)

        masks_b = masks_k[:, None] & masks_bn[None, :]
        b = tl.load(b_ptrs, mask=masks_b, other=0.0)
        b = tl.interleave(b, b)
        b = tl.interleave(b, b)
        b = tl.interleave(b, b)

        # Dequantize b.
        offsets_szk = (
            BLOCK_SIZE_K * SPLIT_K * k + pid_z * BLOCK_SIZE_K
        ) // group_size + tl.arange(0, 1)
        offsets_z = (N // 8) * offsets_szk[:, None] + offsets_zn[None, :]
        masks_zk = offsets_szk < K // group_size
        masks_z = masks_zk[:, None] & masks_zn[None, :]
        zeros_ptrs = zeros_ptr + offsets_z
        zeros = tl.load(zeros_ptrs, mask=masks_z, other=0.0)
        zeros = tl.interleave(zeros, zeros)
        zeros = tl.interleave(zeros, zeros)
        zeros = tl.interleave(zeros, zeros)
        zeros = tl.broadcast_to(zeros, (BLOCK_SIZE_K, BLOCK_SIZE_N))

        offsets_s = N * offsets_szk[:, None] + offsets_sn[None, :]
        masks_sk = offsets_szk < K // group_size
        masks_s = masks_sk[:, None] & masks_sn[None, :]
        scales_ptrs = scales_ptr + offsets_s
        scales = tl.load(scales_ptrs, mask=masks_s, other=0.0)
        scales = tl.broadcast_to(scales, (BLOCK_SIZE_K, BLOCK_SIZE_N))

        b = (b >> shifts) & 0xF
        zeros = (zeros >> shifts) & 0xF
        b = (b - zeros) * scales
        b = b.to(c_ptr.type.element_ty)

        # Accumulate results.
        accumulator = tl.dot(a, b, accumulator, out_dtype=accumulator_dtype)

        offsets_k += BLOCK_SIZE_K * SPLIT_K
        a_ptrs += BLOCK_SIZE_K * SPLIT_K
        b_ptrs += BLOCK_SIZE_K * SPLIT_K * (N // 8)

    c = accumulator.to(c_ptr.type.element_ty)
    offs_cm = pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)
    offs_cn = pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)
    c_ptrs = c_ptr + pid_z * N * M + N * offs_cm[:, None] + offs_cn[None, :]
    c_mask = (offs_cm[:, None] < M) & (offs_cn[None, :] < N)
    tl.store(c_ptrs, c, mask=c_mask)


# qweights - [K     , M // 8], int32
# scales   - [K // G, M     ], float16
# zeros    - [K // G, M // 8], int32
def awq_dequantize_triton(
    qweight: torch.Tensor,
    scales: torch.Tensor,
    zeros: torch.Tensor,
    block_size_x: int = 32,
    block_size_y: int = 32,
) -> torch.Tensor:
    K = qweight.shape[0]
    M = scales.shape[1]
    group_size = qweight.shape[0] // scales.shape[0]

    assert K > 0 and M > 0
    assert scales.shape[0] == K // group_size and scales.shape[1] == M
    assert zeros.shape[0] == K // group_size and zeros.shape[1] == M // 8
    assert group_size <= K
    assert group_size in AWQ_TRITON_SUPPORTED_GROUP_SIZES or group_size == K

    # Result tensor:
    # number of rows = same as input tensor
    # number of cols = 8 x input tensor num cols
    result = torch.empty(
        qweight.shape[0],
        qweight.shape[1] * 8,
        device=qweight.device,
        dtype=scales.dtype,
    )

    Y = qweight.shape[0]  # num rows
    X = qweight.shape[1]  # num cols

    grid = lambda META: (
        triton.cdiv(X, META["BLOCK_SIZE_X"]),
        triton.cdiv(Y, META["BLOCK_SIZE_Y"]),
    )
    awq_dequantize_kernel[grid](
        qweight,
        scales,
        zeros,
        group_size,
        result,
        X,
        Y,
        BLOCK_SIZE_X=block_size_x,
        BLOCK_SIZE_Y=block_size_y,
    )

    return result


# input   - [M, K]
# qweight - [K, N // 8]
# qzeros  - [K // G, N // 8]
# scales  - [K // G, N]
# split_k_iters - parallelism along K-dimension, int, power of 2.
def awq_gemm_triton(
    input: torch.Tensor,
    qweight: torch.Tensor,
    scales: torch.Tensor,
    qzeros: torch.Tensor,
    split_k_iters: int,
    block_size_m: int = 32,
    block_size_n: int = 32,
    block_size_k: int = 32,
) -> torch.Tensor:
    M, K = input.shape
    N = qweight.shape[1] * 8
    group_size = qweight.shape[0] // qzeros.shape[0]

    assert N > 0 and K > 0 and M > 0
    assert qweight.shape[0] == K and qweight.shape[1] == N // 8
    assert qzeros.shape[0] == K // group_size and qzeros.shape[1] == N // 8
    assert scales.shape[0] == K // group_size and scales.shape[1] == N
    assert split_k_iters & (split_k_iters - 1) == 0 and split_k_iters != 0
    assert split_k_iters <= 32
    assert group_size <= K
    assert group_size in AWQ_TRITON_SUPPORTED_GROUP_SIZES or group_size == K

    grid = lambda META: (
        triton.cdiv(M, META["BLOCK_SIZE_M"]) * triton.cdiv(N, META["BLOCK_SIZE_N"]),
        split_k_iters,
    )

    result = torch.zeros((split_k_iters, M, N), dtype=scales.dtype, device=input.device)

    # A = input, B = qweight, C = result
    # A = M x K, B = K x N, C = M x N
    awq_gemm_kernel[grid](
        input,
        qweight,
        result,
        qzeros,
        scales,
        M,
        N,
        K,
        group_size,
        BLOCK_SIZE_M=block_size_m,
        BLOCK_SIZE_N=block_size_n,
        BLOCK_SIZE_K=block_size_k,
        SPLIT_K=split_k_iters,
    )

    result = result.sum(0)

    return result



import math

import torch


def _build_slope_tensor(n_attention_heads: int):
    def get_slopes(n):
        def get_slopes_power_of_2(n):
            start = 2 ** (-(2 ** -(math.log2(n) - 3)))
            ratio = start
            return [start * ratio**i for i in range(n)]

        if math.log2(n).is_integer():
            return get_slopes_power_of_2(
                n
            )  # In the paper, we only train models that have 2^a heads for some a. This function has
        else:  # some good properties that only occur when the input is a power of 2. To maintain that even
            closest_power_of_2 = 2 ** math.floor(
                math.log2(n)
            )  # when the number of heads is not a power of 2, we use this workaround.
            return (
                get_slopes_power_of_2(closest_power_of_2)
                + get_slopes(2 * closest_power_of_2)[0::2][: n - closest_power_of_2]
            )

    # h, 1, 1
    slopes = torch.tensor(get_slopes(n_attention_heads)).reshape(
        n_attention_heads, 1, 1
    )

    return slopes



# Copyright (c) 2023-2025, Songlin Yang, Yu Zhang

import triton
import triton.language as tl

from fla.ops.utils.op import exp, log
from fla.utils import autotune_cache_kwargs


@triton.autotune(
    configs=[
        triton.Config({'BT': BT}, num_warps=num_warps)
        for BT in [16, 32, 64]
        for num_warps in [2, 4, 8]
    ],
    key=['S'],
    **autotune_cache_kwargs,
)
@triton.jit(do_not_specialize=['T'])
def logcumsumexp_fwd_kernel(
    s,
    z,
    T,
    S: tl.constexpr,
    BT: tl.constexpr,
):
    i_bh = tl.program_id(0)
    o_i = tl.arange(0, BT)
    m_s = tl.where(o_i[:, None] >= o_i[None, :], 1., 0.)

    b_mp = tl.full([S], float('-inf'), dtype=tl.float32)
    b_zp = tl.zeros([S], dtype=tl.float32)
    for i_t in range(tl.cdiv(T, BT)):
        p_s = tl.make_block_ptr(s + i_bh * T*S, (T, S), (S, 1), (i_t * BT, 0), (BT, S), (1, 0))
        p_z = tl.make_block_ptr(z + i_bh * T*S, (T, S), (S, 1), (i_t * BT, 0), (BT, S), (1, 0))

        # [BT, S]
        b_s = tl.load(p_s, boundary_check=(0, 1)).to(tl.float32)
        # [S,]
        b_mc = tl.max(b_s, 0)
        b_mc = tl.maximum(b_mp, b_mc)
        b_zp = b_zp * exp(b_mp - b_mc)
        # [BT, S]
        b_s = exp(b_s - b_mc)
        b_z = tl.dot(m_s, b_s, allow_tf32=False) + b_zp
        # [S,]
        b_zc = tl.max(b_z, 0)
        b_mp = b_mc
        b_zp = b_zc
        # [BT, BS]
        # small eps to prevent underflows
        b_z = log(tl.where(b_z != 0, b_z, 1e-20)) + b_mc
        tl.store(p_z, b_z.to(p_z.dtype.element_ty), boundary_check=(0, 1))



import torch
import torch.nn.functional as F
import triton
import triton.language as tl


def naive_softmax(x: torch.Tensor) -> torch.Tensor:
    x_max = x.max(dim=1)[0]
    safe_x = x - x_max[:, None]
    numerator = torch.exp(safe_x)
    denominator = numerator.sum(dim=1)
    softmax_out = numerator / denominator[:, None]
    return softmax_out

def benchmark(func: callable, 
              input_tensor: torch.Tensor, 
              iterations: int = 100
              ) -> float:
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    
    # Warm-up to avoid cold start
    func(input_tensor)
    
    start.record()
    for _ in range(iterations):
        func(input_tensor)
    end.record()
    
    # Wait for all events to complete
    torch.cuda.synchronize()
    
    elapsed_time_ms = start.elapsed_time(end) / iterations
    print(f"Average execution time over {iterations} iterations: {elapsed_time_ms:.3f} ms")
    return elapsed_time_ms


@triton.jit
def _softmax_fwd_kernel(
    output_ptr: tl.tensor,
    stride_output_row: tl.int32,
    input_ptr: tl.tensor,
    stride_input_row: tl.int32,
    num_cols: tl.int32,
    block_size: tl.constexpr,
) -> None:
    row_idx = tl.program_id(0)
    row_start_ptr = input_ptr + (row_idx * stride_input_row) 
    col_offsets = tl.arange(0, block_size)
    input_pointers = row_start_ptr + col_offsets

    row_mask = col_offsets < num_cols

    row = tl.load(input_pointers, mask= row_mask, other=float("-inf"))

    safe_row = row - tl.max(row, axis=0)
    numerator = tl.exp(safe_row)
    denominator = tl.sum(numerator, axis=0)
    softmax_out = numerator / denominator

    output_ptr_row = output_ptr + (row_idx * stride_output_row)
    output_pointers = output_ptr_row + col_offsets
    tl.store(output_pointers, softmax_out, mask=row_mask)

    
def softmax(x: torch.Tensor) -> torch.Tensor:
    """ triton implementation of softmax
    """
    rows, cols = x.shape
    assert x.dim() == 2, f"check tensors are 2D"    
    block_size = triton.next_power_of_2(cols)
    num_warps = 4
    if block_size > 2**11 -1:
        num_warps = 2 * num_warps
    elif block_size > 4095:
        num_warps = 4 * num_warps

    grid = (rows, )
    softmax_out = torch.empty_like(x)

    _softmax_fwd_kernel[grid](
        softmax_out,
        softmax_out.stride(0),
        x,
        x.stride(0),
        cols,
        block_size = block_size,
        num_warps = num_warps
    )
    return softmax_out



if __name__ == "__main__":
    # Run correctness check with smaller sample
    sample = torch.tensor([[1,2,3,4,5], [5,4,3,2,1]], dtype=torch.float32, device="cuda")
    ref_out = F.softmax(sample, dim=1)
    eager_out = naive_softmax(sample)
    triton_out = softmax(sample)

    print(f"{ref_out=}")
    print(f"{eager_out=}")
    print(f"{triton_out=}")
    assert torch.allclose(ref_out, triton_out, atol=1e-6), "The naive softmax does not match PyTorch's softmax!"

    large_sample = torch.randn(10000, 1000, dtype=torch.float32, device="cuda")  # 10k samples, each with 1k features

    # Benchmark PyTorch's softmax and naive softmax
    print("Benchmarking PyTorch's softmax:")
    benchmark(lambda x: F.softmax(x, dim=1), large_sample)

    print("Benchmarking naive softmax:")
    benchmark(naive_softmax, large_sample)

    print("Benchmarking Triton softmax:")
    benchmark(softmax, large_sample)



# Tanh - Pure Triton (fp16 compatible, numerically stable)
# NeuroBrix - NVIDIA Common

import triton
import triton.language as tl


@triton.jit
def _tanh_kernel(x_ptr, out_ptr, n_elements, BLOCK_SIZE: tl.constexpr):
    """Tanh with internal fp32 conversion (no memory copy)."""
    pid = tl.program_id(0)
    offsets = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offsets < n_elements

    x = tl.load(x_ptr + offsets, mask=mask)

    # Convert to fp32 for math
    x_f32 = x.to(tl.float32)

    # Clamp to avoid extreme values
    x_clamped = tl.minimum(tl.maximum(x_f32, -20.0), 20.0)

    # Numerically stable formula
    exp_neg_2x = tl.exp(-2.0 * tl.abs(x_clamped))
    tanh_abs = (1.0 - exp_neg_2x) / (1.0 + exp_neg_2x)

    # Apply sign
    output_f32 = tl.where(x_f32 >= 0.0, tanh_abs, -tanh_abs)

    # Store in original dtype
    tl.store(out_ptr + offsets, output_f32.to(x.dtype), mask=mask)



import torch

import triton
from triton import language as tl


@triton.jit()
def swizzle_tile(tile_id,
                 M, N, K,
                 BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr,
                 GROUP_M: tl.constexpr
                 ):
    grid_m = tl.cdiv(M, BLOCK_M)
    grid_n = tl.cdiv(N, BLOCK_N)
    # re-order program ID for better L2 performance
    width = GROUP_M * grid_n
    group_id = tile_id // width
    group_size = tl.minimum(grid_m - group_id * GROUP_M, GROUP_M)
    pid_m = group_id * GROUP_M + (tile_id % group_size)
    pid_n = (tile_id % width) // group_size
    return pid_m, pid_n


@triton.jit()
def linear_tile(tile_id,
                M, N, K,
                BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr,
                GROUP_M: tl.constexpr
                ):
    pid_m = tile_id // tl.cdiv(N, BLOCK_N)
    pid_n = tile_id % tl.cdiv(N, BLOCK_N)
    return pid_m, pid_n


@triton.jit()
def mac_loop(A, B, C,
             M, N, K,
             locks,
             stride_am, stride_ak, stride_bk, stride_bn, stride_cm, stride_cn,
             iters_per_tile,
             start_iter, end_iter,
             BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr,
             ACC_TYPE: tl.constexpr, GROUP_M: tl.constexpr):

    # where are we in the grid
    tile_id = start_iter // iters_per_tile
    if GROUP_M  > 0:
        pid_m, pid_n = swizzle_tile(tile_id, M, N, K, BLOCK_M, BLOCK_N, BLOCK_K, GROUP_M)
    else:
        pid_m, pid_n = linear_tile(tile_id, M, N, K, BLOCK_M, BLOCK_N, BLOCK_K, GROUP_M)

    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    rk = tl.arange(0, BLOCK_K)
    A = A + (rm[:, None] * stride_am + rk[None, :] * stride_ak) + BLOCK_K * stride_ak * (start_iter % iters_per_tile)
    B = B + (rk[:, None] * stride_bk + rn[None, :] * stride_bn) + BLOCK_K * stride_bk * (start_iter % iters_per_tile)
    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=ACC_TYPE)

    for current_iter in range(start_iter, end_iter):
        a = tl.load(A)
        b = tl.load(B)
        acc += tl.dot(a, b)
        A += BLOCK_K * stride_ak
        B += BLOCK_K * stride_bk

    if end_iter % iters_per_tile == 0:  # last iteration of the tile always happens before its start on another SM
        C_ = C + (rm[:, None] * stride_cm + rn[None, :] * stride_cn)  # compute inside the if/else to avoid spilling!
        tl.store(C_, acc)
        if start_iter % iters_per_tile != 0:  # only if tile has been partially processed
            tl.atomic_xchg(locks + tile_id, 1)
    else:
        while tl.atomic_cas(locks + tile_id, 1, 1) != 1:
            pass
        C_ = C + (rm[:, None] * stride_cm + rn[None, :] * stride_cn)  # compute inside the if/else to avoid spilling!
        tl.atomic_add(C_, acc)


@triton.jit()
def first_wave(
        A, B, C,
        M, N, K,
        locks,
        stride_am, stride_ak, stride_bk, stride_bn, stride_cm, stride_cn,
        total_full_tiles_streamk, total_partial_tiles_streamk, iters_per_tile,
        BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr, ACC_TYPE: tl.constexpr,
        GROUP_M: tl.constexpr,
):
    pid = tl.program_id(0)
    start_iter = pid * total_full_tiles_streamk + tl.minimum(pid, total_partial_tiles_streamk)
    last_iter = (pid + 1) * total_full_tiles_streamk + tl.minimum(pid + 1, total_partial_tiles_streamk)

    while start_iter < last_iter:
        end_iter = tl.minimum(start_iter + (iters_per_tile - start_iter % iters_per_tile), last_iter)
        mac_loop(A, B, C,
                 M, N, K,
                 locks,
                 stride_am, stride_ak, stride_bk, stride_bn, stride_cm, stride_cn,
                 iters_per_tile,
                 start_iter, end_iter,
                 BLOCK_M, BLOCK_N, BLOCK_K, ACC_TYPE,
                 GROUP_M,
                 )

        start_iter = end_iter


@triton.jit()
def full_tiles(
        A, B, C,
        M, N, K,
        stride_am, stride_ak, stride_bk, stride_bn, stride_cm, stride_cn,
        total_tiles_streamk,
        BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr, BLOCK_K: tl.constexpr, ACC_TYPE: tl.constexpr,
        GROUP_M: tl.constexpr,
):
    # first wave has done more tiles than there are SMs, we adjust pid
    tile_id = tl.program_id(0) + total_tiles_streamk
    if GROUP_M > 0:
        pid_m, pid_n = swizzle_tile(tile_id, M, N, K, BLOCK_M, BLOCK_N, BLOCK_K, GROUP_M)
    else:
        pid_m, pid_n = linear_tile(tile_id, M, N, K, BLOCK_M, BLOCK_N, BLOCK_K, GROUP_M)

    # do matrix multiplication
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    rk = tl.arange(0, BLOCK_K)
    # pointers
    A = A + (rm[:, None] * stride_am + rk[None, :] * stride_ak)
    B = B + (rk[:, None] * stride_bk + rn[None, :] * stride_bn)
    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=ACC_TYPE)
    for k in range(0, tl.cdiv(K, BLOCK_K)):
        a = tl.load(A)
        b = tl.load(B)
        acc += tl.dot(a, b)
        A += BLOCK_K * stride_ak
        B += BLOCK_K * stride_bk
    acc = acc.to(tl.float16)  # restore C.dtype.element_ty
    # rematerialize rm and rn to save registers
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    C = C + (rm[:, None] * stride_cm + rn[None, :] * stride_cn)
    tl.store(C, acc)


class matmul(torch.autograd.Function):

    _debug = False

    @staticmethod
    def set_debug(debug: bool):
        matmul._debug = debug

    @staticmethod
    def _call(a: torch.Tensor, b: torch.Tensor, total_programs_streamk: int, BLK_M: int, BLK_N: int, BLK_K: int, two_tiles: bool, num_stages: int, num_warps: int):
        device = a.device

        assert a.is_contiguous() and b.is_contiguous(), "non-contiguous inputs are not supported"
        # checks constraints
        assert a.shape[1] == b.shape[0], "incompatible dimensions"
        M, K = a.shape
        _, N = b.shape
        # accumulator types
        ACC_TYPE = tl.float32 if a.dtype in [torch.float16, torch.bfloat16, torch.float32] else tl.int32
        # compute grid (work to do per SM on the first wave)
        total_blocks_M = triton.cdiv(M, BLK_M)
        total_blocks_N = triton.cdiv(N, BLK_N)
        iters_per_tile = triton.cdiv(K, BLK_K)
        GROUP_M = 8  # 0 to disable swizzling
        total_tiles = total_blocks_M * total_blocks_N

        if total_programs_streamk > 0:  # Stream-K
            # last wave may occupy less than total_programs_streamk SMs
            total_tiles_streamk = total_tiles % total_programs_streamk
            # for two-tile Stream-K + data-parallel from original paper
            if two_tiles and total_tiles - total_tiles_streamk > total_programs_streamk:
                total_tiles_streamk += total_programs_streamk
            # remaining tiles are computed using classical blocking
            total_blocking_tiles = total_tiles - total_tiles_streamk
            total_iters_streamk = total_tiles_streamk * iters_per_tile
            # iterations related to full waves
            total_full_tiles_streamk = total_iters_streamk // total_programs_streamk
            # iterations related to last (partial) wave
            total_partial_tiles_streamk = total_iters_streamk % total_programs_streamk

        else:  # all tiles are computed using classical blocking
            total_blocking_tiles = total_tiles
            total_tiles_streamk = 0
            total_full_tiles_streamk = 0
            total_partial_tiles_streamk = 0
            total_iters_streamk = 0

        if matmul._debug:
            print(f"M,N,K={M},{N},{K} ; BLK_M,N,K={BLK_M},{BLK_N},{BLK_K}")
            print(f"{total_blocks_M=} x {total_blocks_N=} = {total_tiles=}")
            print(f"{total_tiles_streamk=} + {total_blocking_tiles=} = {total_tiles=}")
            print(f"{total_programs_streamk=}")
            print(f"{total_blocking_tiles=}")
            print(f"{iters_per_tile=}")
            print(f"{total_iters_streamk=}")

        # allocates output
        c = torch.empty((M, N), device=device, dtype=a.dtype)
        # allocates locks to sync work accross SMs
        locks = torch.zeros((total_tiles_streamk,), device=device, dtype=torch.int32)
        k1 = first_wave[(total_programs_streamk,)](
            a,
            b,
            c,
            M,
            N,
            K,
            locks,
            a.stride(0),
            a.stride(1),
            b.stride(0),
            b.stride(1),
            c.stride(0),
            c.stride(1),
            total_full_tiles_streamk=total_full_tiles_streamk,
            total_partial_tiles_streamk=total_partial_tiles_streamk,
            iters_per_tile=iters_per_tile,
            BLOCK_M=BLK_M,
            BLOCK_N=BLK_N,
            BLOCK_K=BLK_K,
            ACC_TYPE=ACC_TYPE,
            GROUP_M=GROUP_M,
            num_stages=num_stages,
            num_warps=num_warps,
        )
        if matmul._debug:
            print(f"{k1.n_regs} registers used, {k1.n_spills} spills")
        k2 = full_tiles[(total_blocking_tiles,)](
            a,
            b,
            c,
            M,
            N,
            K,
            a.stride(0),
            a.stride(1),
            b.stride(0),
            b.stride(1),
            c.stride(0),
            c.stride(1),
            total_tiles_streamk=total_tiles_streamk,
            BLOCK_M=BLK_M,
            BLOCK_N=BLK_N,
            BLOCK_K=BLK_K,
            ACC_TYPE=ACC_TYPE,
            GROUP_M=GROUP_M,
            num_stages=num_stages,
            num_warps=num_warps,
        )
        if matmul._debug:
            print(f"{k2.n_regs} registers used, {k2.n_spills} spills")
        return c

    @staticmethod
    def forward(ctx, a: torch.Tensor, b: torch.Tensor, grid: int, BLK_M=128, BLK_N=128, BLK_K=32, two_tiles=True, num_stages=3, num_warps=4):
        return matmul._call(a=a, b=b, total_programs_streamk=grid, BLK_M=BLK_M, BLK_N=BLK_N, BLK_K=BLK_K, two_tiles=two_tiles, num_warps=num_warps, num_stages=num_stages)



import logging

import triton
import triton.language as tl

from flag_gems.utils import tl_extra_shim

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
div_rn = tl_extra_shim.div_rn


@pointwise_dynamic(is_tensor=[True, False], promotion_methods=[(0, "DEFAULT")])
@triton.jit
def silu_forward(x, inplace):
    x_fp32 = x.to(tl.float32)
    y = 1.0 / (1.0 + tl.exp(-x_fp32)) * x_fp32
    return y


@pointwise_dynamic(promotion_methods=[(0, "DEFAULT")])
@triton.jit
def silu_backward_kernel(x, dy):
    dy_fp32 = dy.to(tl.float32)
    x_fp32 = x.to(tl.float32)
    sigma = 1.0 / (1.0 + tl.exp(-x_fp32))
    dx = dy_fp32 * sigma * (1.0 + x_fp32 * (1.0 - sigma))
    return dx


def silu(self):
    logger.debug("GEMS_CAMBRICON SILU FORWARD")
    output = silu_forward(self, False)
    return output


def silu_backward(grad_output, self):
    logger.debug("GEMS_CAMBRICON SILU BACKWARD")
    grad_input = silu_backward_kernel(self, grad_output)
    return grad_input


def silu_(A):
    logger.debug("GEMS_CAMBRICON SILU_ FORWARD")
    out = silu_forward(A, True, out0=A)
    return out



"""
Kernels for gated linear units with arbitrary activation functions.
"""


import triton
import triton.language as tl

from .act_kernels import apply_act_func, apply_act_func_grad
from .utils import element_wise_kernel_configs


@triton.autotune(
    configs=element_wise_kernel_configs(),
    key=['size'],
)
@triton.jit
def glu_forward_kernel(
    input1_pointer, input2_pointer, output_pointer, size, param,
    act_func: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    ):
    """
    Applies the gated linear unit with an arbitrary activation function
    to the input.

    Args:
        input1_pointer: Pointer to the first half of the input to gate.
            The first half must be contiguous and contain size elements.
        input2_pointer: Pointer to the second half of the input to gate.
            The second half must be contiguous and contain size elements.
        output_pointer: Pointer to a container the result is written to.
            The container must be contiguous and contain size elements.
        size: Number of elements in each half of the input.
        param: Parameter in the case of parameterized activation functions.
        act_func: Name of activation function to apply.
            Options are 'sigmoid', 'logsigmoid', 'tanh', 'relu', 'gelu', 'geluapprox', 'silu',
            'relu6', 'hardsigmoid', 'hardtanh', 'hardswish', 'selu', 'mish',
            'softplus', 'softsign', 'tanhshrink', 'leaky_relu', 'elu', 'celu', and 'hardshrink'.
        BLOCK_SIZE: Block size.
    """
    # This program processes BLOCK_SIZE elements.
    pid = tl.program_id(axis=0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offset < size

    input1 = tl.load(input1_pointer + offset, mask=mask)
    input2 = tl.load(input2_pointer + offset, mask=mask)

    output = input1 * apply_act_func(input2, None, None, None, param,
                                     act_func, False)
    tl.store(output_pointer + offset, output, mask=mask)


@triton.autotune(
    configs=element_wise_kernel_configs(),
    key=['size'],
)
@triton.jit
def glu_backward_kernel(
    output_grad_pointer, input1_pointer, input2_pointer,
    input1_grad_pointer, input2_grad_pointer, size, param,
    act_func: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
    ):
    """
    Calculates the input gradient of the gated linear unit.

    Args:
        output_grad_pointer: Pointer to the unit's output gradients.
            The output gradients must be contiguous and contain size elements.
        input1_pointer: Pointer to the first half of the input that was gated.
            The first half must be contiguous and contain size elements.
        input2_pointer: Pointer to the second half of the input that was gated.
            The second half must be contiguous and contain size elements.
        input1_grad_pointer: Pointer to a container the first half's gradients are written to.
            The container must be contiguous and contain size elements.
        input2_grad_pointer: Pointer to a container the second half's gradients are written to.
            The container must be contiguous and contain size elements.
        size: Number of elements in each half of the input.
        param: Parameter in the case of parameterized activation functions.
        act_func: Name of activation function to apply.
            Options are 'sigmoid', 'logsigmoid', 'tanh', 'relu', 'gelu', 'geluapprox', 'silu',
            'softplus', 'softsign', 'tanhshrink', 'leaky_relu', 'elu', 'celu', 'hardshrink',
            and 'softshrink'.
        BLOCK_SIZE: Block size.
    """
    # This program processes BLOCK_SIZE elements.
    pid = tl.program_id(axis=0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    mask = offset < size

    output_grad = tl.load(output_grad_pointer + offset, mask=mask)
    input1 = tl.load(input1_pointer + offset, mask=mask)
    input2 = tl.load(input2_pointer + offset, mask=mask)

    input1_grad = output_grad * apply_act_func(input2, None, None, None, param,
                                               act_func, False)
    input2_grad = output_grad * input1 * apply_act_func_grad(1, input2,
                                                             None, None, None,
                                                             param, act_func,
                                                             False)

    tl.store(input1_grad_pointer + offset, input1_grad, mask=mask)
    tl.store(input2_grad_pointer + offset, input2_grad, mask=mask)



import logging

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import dim_compress, libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


@triton.jit
def welford_func(mean_x, count_x, M_x, mean_y, count_y, M_y):
    count = count_x + count_y
    _count = tl.maximum(count, 1)
    mc_x = mean_x * count_x
    mc_y = mean_y * count_y
    mean = (mc_x + mc_y) / _count
    M = M_x + mc_x * mean_x + M_y + mc_y * mean_y - count * mean * mean
    return mean, count, M


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("var_mean"), key=["M", "N"])
@triton.jit(do_not_specialize=["correction"])
def var_mean_welford_kernel(
    X,
    Var,
    Mean,
    M,
    N,
    correction,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # Map the program id to the row of X it should compute.
    pid = tle.program_id(0) * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    X = X + pid * N
    Var = Var + pid
    Mean = Mean + pid
    row_mask = pid < M

    _mean = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
    _acc = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
    _count = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)

    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask
        x = tl.load(X + cols, mask, other=0.0).to(tl.float32)

        count = _count + mask
        cnt = tl.maximum(count, 1)
        cur_mean = (_mean * _count + x) / cnt
        _acc += (x - cur_mean) * (x - _mean) * mask
        _mean = cur_mean
        _count = count

    # 手动实现 tl.reduce 的功能，沿着 axis=1 进行归约
    # 使用 tl.sum 来进行归约，这等价于 welford 算法在这种情况下的行为

    # 计算每行的总计数
    total_count = tl.sum(_count, axis=1)  # shape: (BLOCK_M,)

    # 计算加权平均值
    weighted_sum = tl.sum(_mean * _count, axis=1)  # shape: (BLOCK_M,)
    mean = weighted_sum / tl.maximum(total_count, 1)  # shape: (BLOCK_M,)

    # 计算方差累积值
    # 对于每个元素，计算其对总体方差的贡献
    mean_expanded = mean[:, None]  # shape: (BLOCK_M, 1)

    # 计算每个局部统计量对总体方差的贡献
    # 这是 Welford 算法的并行化版本
    local_var_contrib = _acc + _count * (_mean - mean_expanded) * (
        _mean - mean_expanded
    )
    acc = tl.sum(local_var_contrib, axis=1)  # shape: (BLOCK_M,)

    var = acc / (N - correction)
    mean = mean[:, None]
    var = var[:, None]

    # Write mean / var
    tl.store(Mean, mean, row_mask)
    tl.store(Var, var, row_mask)


@libentry()
@triton.autotune(configs=runtime.get_tuned_config("var_mean"), key=["M", "N"])
@triton.jit(do_not_specialize=["correction"])
def var_mean_welford_kernel_simple(
    X,
    Var,
    Mean,
    M,
    N,
    correction,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # 程序ID映射
    pid = tle.program_id(0) * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    X = X + pid * N
    Var = Var + pid
    Mean = Mean + pid
    row_mask = pid < M

    # 每行单独处理
    for row in range(BLOCK_M):
        if row < BLOCK_M:
            current_row_mask = (tl.arange(0, BLOCK_M) == row)[:, None] & row_mask

            if tl.sum(current_row_mask.to(tl.int32)) > 0:
                # 初始化当前行的统计量
                running_mean = 0.0
                running_M = 0.0
                count = 0

                # 按块处理当前行
                for off in range(0, N, BLOCK_N):
                    cols = off + tl.arange(0, BLOCK_N)
                    col_mask = cols < N

                    # 加载数据
                    x_vals = tl.load(X + row * N + cols, col_mask, other=0.0).to(
                        tl.float32
                    )

                    # 对块内每个有效元素进行在线更新
                    for i in range(BLOCK_N):
                        if i < BLOCK_N and (off + i) < N:
                            count += 1
                            x = x_vals[i]

                            delta = x - running_mean
                            running_mean += delta / count
                            delta2 = x - running_mean
                            running_M += delta * delta2

                # 计算方差
                variance = running_M / (N - correction) if N > correction else 0.0

                # 存储结果
                tl.store(Mean + row, running_mean, current_row_mask[:, 0])
                tl.store(Var + row, variance, current_row_mask[:, 0])


@libentry()
@triton.jit
def var_mean_kernel_1(
    X,
    Acc,
    Average,
    Count,
    N,
    BLOCK_N: tl.constexpr,
):
    # Map the program id to the row of X it should compute.
    pid = tle.program_id(0)
    offset = pid * BLOCK_N + tl.arange(0, BLOCK_N)

    X = X + offset
    Acc = Acc + pid
    Average = Average + pid
    Count = Count + pid
    mask = offset < N

    x = tl.load(X, mask, other=0.0).to(tl.float32)

    count = tl.sum(mask.to(tl.float32))
    average = tl.sum(x) / count
    acc = tl.sum(x * x) - count * average * average

    tl.store(Average, average)
    tl.store(Acc, acc)
    tl.store(Count, count)


@libentry()
@triton.heuristics(runtime.get_heuristic_config("var_mean"))
@triton.jit(do_not_specialize=["correction"])
def var_mean_kernel_2(
    Acc,
    Average,
    Count,
    Var,
    Mean,
    N,
    correction,
    BLOCK_NUM,
    BLOCK_N: tl.constexpr,
):
    offset = tl.arange(0, BLOCK_N)
    mask = offset < BLOCK_NUM
    Acc = Acc + offset
    Average = Average + offset
    Count = Count + offset
    acc = tl.load(Acc, mask, other=0.0).to(tl.float32)
    average = tl.load(Average, mask, other=0.0).to(tl.float32)
    count = tl.load(Count, mask, other=0.0).to(tl.float32)

    # mean, _, nvar = tl.reduce((average, count, acc), axis=0, combine_fn=welford_func)
    # 手动实现 tl.reduce 的功能，沿着 axis=0 进行归约
    # 计算总计数
    total_count = tl.sum(count)

    # 计算加权平均值
    weighted_sum = tl.sum(average * count)
    mean = weighted_sum / tl.maximum(total_count, 1)

    # 计算方差累积值
    # 对于每个块，计算其对总体方差的贡献
    # 这是 Welford 算法的并行化版本
    local_var_contrib = acc + count * (average - mean) * (average - mean)
    nvar = tl.sum(local_var_contrib)

    var = nvar / (N - correction)
    tl.store(Mean, mean)
    tl.store(Var, var)


def var_mean(x, dim=None, *, correction=None, keepdim=False):
    logger.debug("GEMS_ASCEND VAR MEAN")
    if correction is None:
        correction = 1.0

    if dim is None or len(dim) == x.ndim:
        dim = list(range(x.ndim))
        shape = [1] * x.ndim
        N = x.numel()
        var = torch.empty(shape, dtype=x.dtype, device=x.device)
        mean = torch.empty(shape, dtype=x.dtype, device=x.device)
        BLOCK_N = 1024
        BLOCK_NUM = triton.cdiv(N, BLOCK_N)
        acc = torch.empty([BLOCK_NUM], dtype=x.dtype, device=x.device)
        average = torch.empty([BLOCK_NUM], dtype=x.dtype, device=x.device)
        count = torch.empty([BLOCK_NUM], dtype=x.dtype, device=x.device)

        with torch_device_fn.device(x.device):
            var_mean_kernel_1[(BLOCK_NUM,)](x, acc, average, count, N, BLOCK_N=BLOCK_N)
            var_mean_kernel_2[(1,)](
                acc, average, count, var, mean, N, correction, BLOCK_NUM
            )
    else:
        shape = list(x.shape)
        dim = [d % x.ndim for d in dim]
        x = dim_compress(x, dim)
        N = 1
        for i in dim:
            N *= shape[i]
            shape[i] = 1
        M = x.numel() // N
        var = torch.empty(shape, dtype=x.dtype, device=x.device)
        mean = torch.empty(shape, dtype=x.dtype, device=x.device)

        grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]),)
        with torch_device_fn.device(x.device):
            var_mean_welford_kernel[grid](x, var, mean, M, N, correction)

    if not keepdim:
        var = var.squeeze(dim=dim)
        mean = mean.squeeze(dim=dim)
    return var, mean



import logging

import torch
import triton
import triton.language as tl

from flag_gems.utils import tl_extra_shim

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))
_isfinited = tl_extra_shim.isfinited
_finitef = tl_extra_shim.finitef


@pointwise_dynamic(is_tensor=[True], promotion_methods=[(0, "ALWAYS_BOOL")])
@triton.jit
def isfinite_func(x):
    return _isfinited(x) if x.dtype.is_fp64() else _finitef(x.to(tl.float32))


def isfinite(
    A: torch.Tensor,
) -> torch.Tensor:
    logger.debug("GEMS ISFINITE")
    if A.is_floating_point():
        return isfinite_func(A)
    else:
        return torch.full(A.shape, True, dtype=torch.bool, device=A.device)



import logging
import math

import torch
import triton
import triton.language as tl

from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import libentry
from flag_gems.utils import triton_lang_extension as tle
from flag_gems.utils.limits import get_dtype_min

logger = logging.getLogger(f'flag_gems.runtime._ascend.ops.{__name__.split(".")[-1]}')


@libentry()
@triton.jit
def argmax_kernel_1(
    inp,
    mid_value,
    mid_index,
    M,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < M
    min_value = get_dtype_min(inp.type.element_ty)
    inp_val = tl.load(inp_ptrs, mask=mask, other=min_value)
    max_val, max_index = tl.max(inp_val, axis=0, return_indices=True)
    max_index = max_index + pid * BLOCK_SIZE
    mid_value_ptr = mid_value + pid
    max_index_ptr = mid_index + pid
    tl.store(mid_value_ptr, max_val)
    tl.store(max_index_ptr, max_index)


@libentry()
@triton.jit
def argmax_kernel_2(mid_value, mid_index, out, mid_size, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid_value + offset
    mask = offset < mid_size
    min_value = get_dtype_min(mid_value.type.element_ty)
    mid_val = tl.load(mid_ptrs, mask=mask, other=min_value)
    index_val = tl.argmax(mid_val, axis=0)
    mid_index_ptrs = mid_index + index_val
    out_val = tl.load(mid_index_ptrs)
    tl.store(out, out_val)


@libentry()
@triton.heuristics(runtime.get_heuristic_config("argmax"))
@triton.jit
def argmax_kernel(
    inp,
    out_index,
    M,
    N,
    K,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
):
    # set offset
    pid_m = tle.program_id(0)
    # pid_k = tle.program_id(1)
    for pid_k in range(K):
        m_offset = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)

        dtype = inp.type.element_ty
        acc_type = tl.float32 if dtype is tl.bfloat16 else dtype
        min_value = get_dtype_min(dtype)
        max_values = tl.full([BLOCK_M], dtype=acc_type, value=min_value)
        argmax_values = tl.full([BLOCK_M], dtype=tl.int64, value=0)
        for start_n in range(0, N, BLOCK_N):
            n_offset = start_n + tl.arange(0, BLOCK_N)
            offset = m_offset[:, None] * N * K + n_offset[None, :] * K + pid_k
            mask = m_offset[:, None] < M and n_offset[None, :] < N
            inp_ptrs = inp + offset
            inp_vals = tl.load(inp_ptrs, mask=mask, other=min_value)
            local_max, local_argmax = tl.max(
                inp_vals, 1, return_indices=True, return_indices_tie_break_left=True
            )
            # if return indices is not supported, call a tl.argmax in addition
            # local_argmax = tl.argmax(inp_vals, 1)
            update = local_max > max_values
            max_values = tl.where(update, local_max, max_values)
            argmax_values = tl.where(update, start_n + local_argmax, argmax_values)

        offset_index = m_offset * K + pid_k
        out_index_ptrs = out_index + offset_index
        mask1 = m_offset < M
        tl.store(out_index_ptrs, argmax_values, mask=mask1)


def argmax(inp, dim=None, keepdim=False, *, dtype=None):
    logger.debug("GEMS_ASCEND ARGMAX")
    if dim is None:
        M = inp.numel()
        if dtype is None:
            dtype = inp.dtype
        block_size = triton.next_power_of_2(math.ceil(math.sqrt(M)))
        mid_size = triton.cdiv(M, block_size)
        block_mid = triton.next_power_of_2(mid_size)

        mid_value = torch.empty((mid_size,), dtype=dtype, device=inp.device)
        mid_index = torch.empty((mid_size,), dtype=torch.int64, device=inp.device)
        if keepdim:
            shape = list(inp.shape)
            for i in range(0, inp.dim()):
                shape[i] = 1
            out = torch.empty(shape, dtype=torch.int64, device=inp.device)
        else:
            out = torch.empty([], dtype=torch.int64, device=inp.device)

        with torch_device_fn.device(inp.device):
            argmax_kernel_1[(mid_size, 1, 1)](
                inp,
                mid_value,
                mid_index,
                M,
                block_size,
            )
            argmax_kernel_2[(1, 1, 1)](mid_value, mid_index, out, mid_size, block_mid)
        return out
    else:
        assert dim >= -inp.ndim and dim < inp.ndim, "Invalid dim"
        shape = inp.shape
        dim = dim % inp.ndim
        N = shape[dim]
        M = math.prod(shape[:dim])
        K = inp.numel() // M // N

        inp = inp.contiguous()

        shape_list = list(shape)
        shape_list[dim] = 1
        out_index = torch.empty(shape_list, dtype=torch.int64, device=inp.device)
        if not keepdim:
            out_index = torch.squeeze(out_index, dim)
        grid = lambda meta: (
            triton.cdiv(M, meta["BLOCK_M"]),
            # K,
        )
        with torch_device_fn.device(inp.device):
            argmax_kernel[grid](
                inp,
                out_index,
                M,
                N,
                K,
            )

        return out_index



"""
NeuroBrix Validators - Operation Validator.

Validates that all operations in graph are supported.

ZERO FALLBACK: Unsupported operations raise explicit UnsupportedOpsError.
"""

from typing import Any, Dict, List, Set

from .base import BaseValidator, SafetyLevel, ValidationResult, UnsupportedOpsError
from .config import get_config


class OpValidator(BaseValidator):
    """
    Validates operation support.

    Checks:
    - All ops are in supported set
    - Op parameters are valid
    - Op combinations are allowed
    - Hardware-specific op availability
    """

    @property
    def name(self) -> str:
        return "OpValidator"

    def __init__(self, safety_level: SafetyLevel = SafetyLevel.STANDARD):
        super().__init__(safety_level)
        self.config = get_config()

    def validate(self, graph: Any, context: Dict[str, Any]) -> ValidationResult:
        """
        Validate all operations in the graph.

        Args:
            graph: Execution graph with nodes
            context: Optional:
                - 'hardware': Hardware type for availability check
                - 'custom_ops': Additional supported ops

        Returns:
            ValidationResult with op validation status
        """
        warnings = []
        errors = []
        metrics = {}

        # Get supported ops (config + custom)
        supported_ops = set(self.config.supported_ops)
        custom_ops = context.get('custom_ops', set())
        supported_ops.update(custom_ops)

        # Hardware-specific ops
        hardware = context.get('hardware', 'generic')
        hw_ops = self._get_hardware_ops(hardware)
        supported_ops.update(hw_ops)

        # Collect all ops from graph
        found_ops: Set[str] = set()
        unsupported: List[str] = []
        op_counts: Dict[str, int] = {}

        nodes = getattr(graph, 'nodes', [])
        if hasattr(graph, '__iter__') and not nodes:
            nodes = list(graph)

        for node in nodes:
            # Support both attribute and dict access (ONNX format)
            if hasattr(node, 'op'):
                op = node.op
            elif hasattr(node, 'op_type'):
                op = node.op_type
            elif isinstance(node, dict):
                op = node.get('op_type') or node.get('op')
            else:
                op = None

            if op is None:
                continue

            op_lower = op.lower()
            found_ops.add(op_lower)
            op_counts[op_lower] = op_counts.get(op_lower, 0) + 1

            if op_lower not in supported_ops:
                unsupported.append(op)

            # Validate op parameters
            self._validate_op_params(node, op_lower, warnings, errors)

        # Remove duplicates from unsupported
        unsupported_unique = list(set(unsupported))

        metrics['total_ops'] = len(nodes)
        metrics['unique_ops'] = len(found_ops)
        metrics['op_distribution'] = op_counts

        if unsupported_unique:
            errors.append(
                f"Found {len(unsupported_unique)} unsupported operation(s): "
                f"{', '.join(sorted(unsupported_unique))}"
            )
            metrics['unsupported_ops'] = unsupported_unique

        if errors:
            return ValidationResult(
                passed=False,
                validator_name=self.name,
                message="\n".join(errors),
                warnings=warnings,
                metrics=metrics,
            )

        return ValidationResult(
            passed=True,
            validator_name=self.name,
            message=f"All {len(found_ops)} unique operations are supported",
            warnings=warnings,
            metrics=metrics,
        )

    def _get_hardware_ops(self, hardware: str) -> Set[str]:
        """Get hardware-specific operations."""
        hw_ops: Dict[str, Set[str]] = {
            'nvidia': {
                'flash_attention_2',
                'tensorcore_matmul',
                'cudnn_conv',
            },
            'amd': {
                'rocm_attention',
                'hipblas_matmul',
            },
            'apple': {
                'mps_conv',
                'ane_matmul',
            },
        }
        return hw_ops.get(hardware.lower(), set())

    def _validate_op_params(
        self,
        node: Any,
        op: str,
        warnings: List[str],
        errors: List[str],
    ) -> None:
        """Validate operation-specific parameters."""
        # Matmul/Linear checks
        if op in ('matmul', 'linear'):
            # Check for transpose flags if present
            trans_a = getattr(node, 'transpose_a', False)
            trans_b = getattr(node, 'transpose_b', False)
            # Both transposes are valid, no error

        # Convolution checks
        elif op in ('conv2d', 'conv1d'):
            stride = getattr(node, 'stride', 1)
            padding = getattr(node, 'padding', 0)
            dilation = getattr(node, 'dilation', 1)
            groups = getattr(node, 'groups', 1)

            if isinstance(stride, int) and stride <= 0:
                errors.append(f"Invalid stride {stride} for {op}")
            if isinstance(dilation, int) and dilation <= 0:
                errors.append(f"Invalid dilation {dilation} for {op}")
            if isinstance(groups, int) and groups <= 0:
                errors.append(f"Invalid groups {groups} for {op}")

        # Attention checks
        elif op in ('scaled_dot_product_attention', 'flash_attention'):
            dropout_p = getattr(node, 'dropout_p', 0.0)
            if dropout_p < 0 or dropout_p > 1:
                errors.append(f"Invalid dropout_p {dropout_p} for {op}")

            scale = getattr(node, 'scale', None)
            if scale is not None and scale <= 0:
                errors.append(f"Invalid scale {scale} for {op}")

        # Normalization checks
        elif op in ('layernorm', 'batchnorm', 'groupnorm'):
            eps = getattr(node, 'eps', 1e-5)
            if eps <= 0:
                errors.append(f"Invalid eps {eps} for {op}")

            if op == 'groupnorm':
                num_groups = getattr(node, 'num_groups', None)
                if num_groups is not None and num_groups <= 0:
                    errors.append(f"Invalid num_groups {num_groups} for groupnorm")

        # Reduction checks
        elif op in ('sum', 'mean', 'max', 'min'):
            keepdim = getattr(node, 'keepdim', False)
            # keepdim is valid as bool, no error

        # Activation checks
        elif op == 'leaky_relu':
            negative_slope = getattr(node, 'negative_slope', 0.01)
            if negative_slope < 0:
                warnings.append(f"Unusual negative_slope {negative_slope} for leaky_relu")

        # Pooling checks
        elif op in ('maxpool2d', 'avgpool2d'):
            kernel_size = getattr(node, 'kernel_size', None)
            if kernel_size is not None:
                if isinstance(kernel_size, int) and kernel_size <= 0:
                    errors.append(f"Invalid kernel_size {kernel_size} for {op}")

    def check_op_supported(self, op: str, hardware: str = 'generic') -> bool:
        """
        Check if an operation is supported.

        Args:
            op: Operation name
            hardware: Target hardware

        Returns:
            True if supported
        """
        supported = set(self.config.supported_ops)
        supported.update(self._get_hardware_ops(hardware))
        return op.lower() in supported

    def get_all_supported_ops(self, hardware: str = 'generic') -> Set[str]:
        """Get set of all supported operations."""
        supported = set(self.config.supported_ops)
        supported.update(self._get_hardware_ops(hardware))
        return supported

    def raise_if_unsupported(
        self,
        ops: List[str],
        hardware: str = 'generic',
    ) -> None:
        """
        Raise exception if any ops are unsupported.

        Args:
            ops: List of operation names
            hardware: Target hardware

        Raises:
            UnsupportedOpsError if any ops unsupported
        """
        supported = self.get_all_supported_ops(hardware)
        unsupported = [op for op in ops if op.lower() not in supported]

        if unsupported:
            raise UnsupportedOpsError(
                f"Unsupported operations: {', '.join(unsupported)}",
                unsupported_ops=unsupported,
                suggestion=f"Available ops: {', '.join(sorted(supported)[:20])}... "
                          f"Consider implementing custom kernels or using decomposition.",
            )

    def should_run(self, safety_level: SafetyLevel) -> bool:
        """Op validation runs at all levels."""
        return True



#  Copyright 2022 Lefebvre Sarrut
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#
# This source code is licensed under the BSD license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import torch
import triton
import triton.language as tl
from torch.autograd.function import FunctionCtx
from torch.cuda.amp import custom_fwd
from triton.ops.matmul_perf_model import early_config_prune, estimate_matmul_time

from kernl.implementations import activation_func


# CREDITS: Initially inspired by the Triton tutorial on matrix multiplications


def init_to_zero(name):
    return lambda nargs: nargs[name].zero_()


def get_configs_io_bound():
    configs = []
    for num_stages in [2, 3, 4, 5, 6]:
        for block_m in [16, 32]:
            for block_k in [32, 64]:
                for block_n in [32, 64, 128, 256]:
                    num_warps = 2 if block_n <= 64 else 4
                    configs.append(
                        triton.Config(
                            {"BLOCK_M": block_m, "BLOCK_N": block_n, "BLOCK_K": block_k, "SPLIT_K": 1},
                            num_stages=num_stages,
                            num_warps=num_warps,
                        )
                    )
                    # split_k not used
                    # for split_k in [2, 4, 8, 16]:
                    #     configs.append(triton.Config(
                    #         {'BLOCK_M': block_m, 'BLOCK_N': block_n, 'BLOCK_K': block_k, 'SPLIT_K': split_k},
                    #         num_stages=num_stages, num_warps=num_warps, pre_hook=init_to_zero('C')))
    return configs


@triton.autotune(
    configs=[
        triton.Config({"BLOCK_M": 128, "BLOCK_N": 256, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=3, num_warps=8),
        triton.Config({"BLOCK_M": 256, "BLOCK_N": 128, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=3, num_warps=8),
        triton.Config({"BLOCK_M": 256, "BLOCK_N": 64, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 64, "BLOCK_N": 256, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 128, "BLOCK_N": 128, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 128, "BLOCK_N": 64, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 64, "BLOCK_N": 128, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 128, "BLOCK_N": 32, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 64, "BLOCK_N": 32, "BLOCK_K": 32, "SPLIT_K": 1}, num_stages=5, num_warps=2),
        # good for int8
        triton.Config({"BLOCK_M": 128, "BLOCK_N": 256, "BLOCK_K": 128, "SPLIT_K": 1}, num_stages=3, num_warps=8),
        triton.Config({"BLOCK_M": 256, "BLOCK_N": 128, "BLOCK_K": 128, "SPLIT_K": 1}, num_stages=3, num_warps=8),
        triton.Config({"BLOCK_M": 256, "BLOCK_N": 64, "BLOCK_K": 128, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 64, "BLOCK_N": 256, "BLOCK_K": 128, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 128, "BLOCK_N": 128, "BLOCK_K": 128, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 128, "BLOCK_N": 64, "BLOCK_K": 64, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 64, "BLOCK_N": 128, "BLOCK_K": 64, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 128, "BLOCK_N": 32, "BLOCK_K": 64, "SPLIT_K": 1}, num_stages=4, num_warps=4),
        triton.Config({"BLOCK_M": 64, "BLOCK_N": 32, "BLOCK_K": 64, "SPLIT_K": 1}, num_stages=5, num_warps=2),
    ]
    + get_configs_io_bound(),
    key=["CACHE_KEY_M", "CACHE_KEY_N", "CACHE_KEY_K"],
    prune_configs_by={"early_config_prune": early_config_prune, "perf_model": estimate_matmul_time, "top_k": 10},
)
@triton.heuristics(
    {
        "K_LOAD_MASK_NEEDED": lambda args: args["K"] % (args["BLOCK_K"] * args["SPLIT_K"]) == 0,
    }
)
@triton.jit
def kernel_fma(
    C,  # Pointers to matrices
    ACT_INPUTS,
    A,
    B,
    bias,
    # Matrix dimensions
    M,
    N,
    K,
    CACHE_KEY_M,
    CACHE_KEY_N,
    CACHE_KEY_K,
    # The stride variables represent how much to increase the ptr by when moving by 1
    # element in a particular dimension. E.g. stride_am is how much to increase a_ptr
    # by to get the element one row down (A has M rows)
    output_m_stride,
    output_n_stride,
    act_inputs_m_stride,
    act_inputs_n_stride,
    a_m_stride,
    a_k_stride,
    b_n_stride,
    b_k_stride,
    # Meta-parameters
    BLOCK_M: tl.constexpr,
    GROUP_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_K: tl.constexpr,
    # split k not used, not performant with activation, kept because early_config_prune is expecting it
    SPLIT_K: tl.constexpr,
    K_LOAD_MASK_NEEDED: tl.constexpr,
    HAS_BIAS: tl.constexpr,
    SHOULD_SAVE_ACT_INPUTS: tl.constexpr,
    ACTIVATION: tl.constexpr,
):
    """
    Kernel for computing Out = activation(A x W + C)

    - Input has shape (M, K)
    - Weight has shape (K, N)
    - Bias has shape (N,)
    - Output has shape (M, N)
    - ActInputs (optional) has shape (M, N)

    'ActInputs' optionally saves the A x W + C intermediate for backward computations

    This kernel will consolidate over K
    """
    program_idx = tl.program_id(axis=0)

    grid_m = (M + BLOCK_M - 1) // BLOCK_M
    grid_n = (N + BLOCK_N - 1) // BLOCK_N
    # re-order program ID for better L2 performance
    width = GROUP_M * grid_n
    group_idx = program_idx // width
    group_size = min(grid_m - group_idx * GROUP_M, GROUP_M)
    block_m_idx = group_idx * GROUP_M + (program_idx % group_size)
    block_n_idx = (program_idx % width) // group_size

    # now compute the block that each program will go through
    # m_offs (resp. n_offs) denotes a range of indices
    # for rows (resp. col) of C
    m_offs_untagged = block_m_idx * BLOCK_M + tl.arange(0, BLOCK_M)
    n_offs_untagged = block_n_idx * BLOCK_N + tl.arange(0, BLOCK_N)

    # trick to avoid masking on M and N axis
    # m_offs_untagged and n_offs_untagged can contains addresses outside matrix boundaries
    # modulo operation is used to wrap around the indices that go beyond the matrix boundaries
    # The value loaded are not ok but at least we are not reading outside the A/B matrices
    # Then, during storing in C a mask is used and the results related to these wrong values is discarded!
    # Regarding max_contiguous and multiple_of, they are used to force the compiler to vectorize loads
    # multiple_of indicates that the first element of rm / rn is a multiple of BLOCK_M / BLOCK_N
    # max_contiguous indicates that the range is a block of BLOCK_M / BLOCK_N contiguous elements
    m_offs = tl.max_contiguous(tl.multiple_of(m_offs_untagged % M, BLOCK_M), BLOCK_M)
    n_offs = tl.max_contiguous(tl.multiple_of(n_offs_untagged % N, BLOCK_N), BLOCK_N)

    k_range_offs = tl.arange(0, BLOCK_K)

    A = A + (m_offs[:, None] * a_m_stride + k_range_offs[None, :] * a_k_stride)
    B = B + (k_range_offs[:, None] * b_k_stride + n_offs[None, :] * b_n_stride)

    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)

    if HAS_BIAS:
        bias = tl.load(bias + n_offs, mask=n_offs < N, other=0.0).to(tl.float32)
        acc += bias[None, :]

    for k in range(K, 0, -BLOCK_K):
        if K_LOAD_MASK_NEEDED:
            a = tl.load(A)
            b = tl.load(B)
        else:
            a = tl.load(A, mask=k_range_offs[None, :] < k, other=0.0)
            b = tl.load(B, mask=k_range_offs[:, None] < k, other=0.0)
        acc += tl.dot(a, b)

        A += BLOCK_K * a_k_stride
        B += BLOCK_K * b_k_stride

    # optional: save the activation inputs
    if SHOULD_SAVE_ACT_INPUTS:
        act_in_ptrs = ACT_INPUTS + m_offs[:, None] * act_inputs_m_stride + n_offs[None, :] * act_inputs_n_stride
        tl.store(act_in_ptrs, acc)

    # optional: fused activation (while the data is in shared memory)
    if ACTIVATION == "tanh":
        acc = activation_func.tanh(acc)
    if ACTIVATION == "gelu":
        acc = activation_func.gelu(acc)
    if ACTIVATION == "fast_gelu":
        acc = activation_func.fast_gelu(acc)
    if ACTIVATION == "relu":
        acc = activation_func.relu(acc)

    # write back result
    C = C + m_offs[:, None] * output_m_stride + n_offs[None, :] * output_n_stride
    c_ptr_mask = (m_offs < M)[:, None] & (n_offs < N)[None, :]
    tl.store(C, acc, mask=c_ptr_mask)


class LinearLayer(torch.autograd.Function):
    @staticmethod
    @custom_fwd(cast_inputs=torch.float16)
    def forward(
        ctx: FunctionCtx,
        x: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor],
        activation: str,
        act_inputs: Optional[torch.Tensor],
    ) -> torch.Tensor:
        """
        Compute e = activation(x @ weight + bias).
        This wrapper kicks the `kernel_fma` Triton kernel
        :param ctx: context for autograd
        :param x: input tensor
        :param weight: weight matrix
        :param bias: an optional bias tensor
        :param activation: Activation name. Needs to be a Triton kernel.
        :param act_inputs: an optional tensor to save the activation inputs (for backward)
        :return: result tensor
        """
        x_ = x if x.ndim == 2 else x.flatten(0, 1)

        assert x.dtype == weight.dtype, f"Input and weight must have the same dtype, got {x.dtype} and {weight.dtype}"
        if bias is not None:
            assert x.dtype == bias.dtype, f"Input and bias must have the same dtype, got {x.dtype} and {bias.dtype}"
        assert x_.shape[1] == weight.shape[1], f"Incompatible dimensions: {x_.shape} - {weight.shape}"

        assert bias is None or bias.is_contiguous()
        assert bias is None or bias.shape[0] == weight.shape[0], "Incompatible dimensions in between weight and bias"
        assert weight.is_contiguous()

        M, K = x_.shape
        N, K = weight.shape

        outputs = torch.empty((M, N), device=x.device, dtype=x.dtype)

        # 1D launch kernel where each block gets its own program.
        grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]) * triton.cdiv(N, META["BLOCK_N"]),)  # noqa

        kernel_fma[grid](
            outputs,
            act_inputs,
            x_,
            weight,  # data ptrs
            bias if bias is not None else x,  # auto skip bias if not present
            M,  # shapes
            N,
            K,
            M // 32,  # key for triton cache (limit number of compilations)
            N // 32,
            K // 32,
            output_m_stride=outputs.stride(0),  # strides
            output_n_stride=outputs.stride(1),
            act_inputs_m_stride=act_inputs.stride(0) if act_inputs is not None else 0,
            act_inputs_n_stride=act_inputs.stride(1) if act_inputs is not None else 0,
            a_m_stride=x_.stride(0),
            a_k_stride=x_.stride(1),
            b_n_stride=weight.stride(0),
            b_k_stride=weight.stride(1),
            HAS_BIAS=bias is not None,  # optional fused bias
            SHOULD_SAVE_ACT_INPUTS=act_inputs is not None,  # optional save activation inputs
            ACTIVATION=activation if not None else x,  # optional fused activation
            GROUP_M=8,  # speed optimization: group the programs
        )

        outputs = outputs if x.ndim == 2 else outputs.reshape(x.shape[0], -1, N)
        ctx.save_for_backward(weight, bias, x)
        return outputs


def linear_layer(
    x: torch.Tensor,
    weight: torch.Tensor,
    bias: Optional[torch.Tensor],
    activation="",
    act_inputs: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    return LinearLayer.apply(x, weight, bias, activation, act_inputs)



import logging

import triton
import triton.language as tl
import triton.language.extra.xpu.libdevice as xpu

from ..utils.pointwise_dynamic import pointwise_dynamic

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@pointwise_dynamic(promotion_methods=[(0, "INT_TO_FLOAT")])
@triton.jit
def rsqrt_func(x):
    return xpu.rsqrt(x.to(tl.float32))


def rsqrt(A):
    logger.debug("GEMS RSQRT")
    return rsqrt_func(A)


def rsqrt_(A):
    logger.debug("GEMS RSQRT_")
    return rsqrt_func(A, out0=A)



# This custom op requires musa device capability >= 31.
# We determine whether to enable this op by distinguish the op registration for different arch.

import logging

import torch
import triton
import triton.language as tl

from flag_gems.utils import pointwise_dynamic, tl_extra_shim

pow = tl_extra_shim.pow
fast_tanh = tl_extra_shim.fast_tanh

logger = logging.getLogger(
    f'flag_gems.runtime.backend._mthreads.ops.{__name__.split(".")[-1]}'
)


@pointwise_dynamic(promotion_methods=[(0, "INT_TO_FLOAT")])
@triton.jit
def tanh_forward(x):
    return fast_tanh(x.to(tl.float32))


@pointwise_dynamic(promotion_methods=[(0, "INT_TO_FLOAT")])
@triton.jit
def tanh_backward(y, dy):
    return dy * (1.0 - pow(y.to(tl.float32), 2))


class Tanh(torch.autograd.Function):
    @staticmethod
    def forward(ctx, A):
        logger.debug("GEMS_MTHREADS TANH FORWARD")
        if A.requires_grad is True:
            out = tanh_forward(A.to(torch.float32))
            ctx.save_for_backward(out)
            return out.to(A.dtype)
        else:
            out = tanh_forward(A)
            return out

    @staticmethod
    def backward(ctx, out_grad):
        logger.debug("GEMS_MTHREADS TANH BACKWARD")
        (out,) = ctx.saved_tensors
        in_grad = tanh_backward(out, out_grad)
        return in_grad


def tanh(A):
    return Tanh.apply(A)



"""
Reshape - Reshape tensor to new shape
ATen: aten::reshape, aten::view
"""
import torch
from typing import List

def reshape_kernel(x: torch.Tensor, shape: list = None, size: list = None) -> torch.Tensor:
    """Reshape tensor."""
    target_shape = shape if shape is not None else size
    if target_shape is None:
        return x
    return x.reshape(*target_shape)



import logging

import torch
import triton
import triton.language as tl

from flag_gems.utils import libentry
from flag_gems.utils import triton_lang_extension as tle

logger = logging.getLogger(__name__)


@triton.jit
def prev_multiple_of(a, b):
    # the largest x<a that x%b ==0
    return tl.cdiv(a, b) * b - b


@triton.jit()
def swizzle_tile(
    tile_id,
    M,
    N,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    GROUP_M: tl.constexpr,
):
    grid_m = tl.cdiv(M, BLOCK_M)
    grid_n = tl.cdiv(N, BLOCK_N)
    # re-order program ID for better L2 performance
    width = GROUP_M * grid_n
    group_id = tile_id // width
    group_size = tl.minimum(grid_m - group_id * GROUP_M, GROUP_M)
    pid_m = group_id * GROUP_M + (tile_id % group_size)
    pid_n = (tile_id % width) // group_size
    return pid_m, pid_n


@triton.jit()
def linear_tile(
    tile_id,
    M,
    N,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    GROUP_M: tl.constexpr,
):
    grid_n = tl.cdiv(N, BLOCK_N)

    # column first
    pid_m = tile_id // grid_n
    pid_n = tile_id % grid_n

    return pid_m, pid_n


@triton.jit(
    do_not_specialize=[
        "iters_per_pid",
        "iters_remaining",
        "iters_per_tile",
        "start_iter",
        "end_iter",
    ]
)
def mac_loop(
    A,
    B,
    C,
    P,
    M,
    N,
    K,
    locks,
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    iters_per_pid,
    iters_remaining,
    iters_per_tile,
    start_iter,
    end_iter,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_K: tl.constexpr,
    GROUP_M: tl.constexpr,
):
    # where are we in the grid
    pid = tle.program_id(0)
    tile_id = start_iter // iters_per_tile

    pid_m, pid_n = swizzle_tile(tile_id, M, N, BLOCK_M, BLOCK_N, GROUP_M)

    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    rk = tl.arange(0, BLOCK_K)

    if stride_am == 1:
        ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
    else:
        ram = rm % M
    if stride_bk == 1:
        rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)
    else:
        rbn = rn % N

    # pointers
    A_base = A + ram[:, None] * stride_am
    B_base = B + rbn[None, :] * stride_bn
    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)

    if end_iter % iters_per_tile != 0:
        for current_iter in range(start_iter, end_iter):
            k_offset_in_tile = (current_iter % iters_per_tile) * BLOCK_K
            a = tl.load(A_base + (k_offset_in_tile + rk[None, :]) * stride_ak)
            b = tl.load(B_base + (k_offset_in_tile + rk[:, None]) * stride_bk)
            acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)
    else:
        prev_multiple = prev_multiple_of(K, BLOCK_K)
        for current_iter in range(start_iter, end_iter - 1):
            k_offset_in_tile = (current_iter % iters_per_tile) * BLOCK_K
            a = tl.load(A_base + (k_offset_in_tile + rk[None, :]) * stride_ak)
            b = tl.load(B_base + (k_offset_in_tile + rk[:, None]) * stride_bk)
            acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)

        # handle the last iter
        rk = prev_multiple + tl.arange(0, BLOCK_K)
        mask_k = rk < K
        a = tl.load(A_base + rk[None, :] * stride_ak, mask=mask_k[None, :])
        b = tl.load(B_base + rk[:, None] * stride_bk, mask=mask_k[:, None])
        acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)

    rm1 = tl.arange(0, BLOCK_M)
    rn1 = tl.arange(0, BLOCK_N)

    # the first situation: not the starting parts. only need to store the data on P
    if start_iter % iters_per_tile != 0:
        P_ptr = P + pid * BLOCK_M * BLOCK_N + (rm1[:, None] * BLOCK_N + rn1[None, :])
        tl.store(P_ptr, acc, cache_modifier=".cg")
        # tl.debug_barrier()
        tl.atomic_xchg(locks + pid, 1)
    else:  # the first part of certain grids. shoud read datas and merge datas
        next_pid = pid + 1
        stop_loading_iter = start_iter + iters_per_tile
        end = end_iter
        while end < stop_loading_iter:
            while tl.atomic_cas(locks + next_pid, 1, 1) != 1:
                pass
            P_ptr = (
                P
                + next_pid * BLOCK_M * BLOCK_N
                + (rm1[:, None] * BLOCK_N + rn1[None, :])
            )
            acc += tl.load(P_ptr, cache_modifier=".cg")
            end += iters_per_pid + (next_pid < iters_remaining)
            next_pid += 1

        # acc = acc.to(C.dtype.element_ty)  #
        C_ = C + (rm[:, None] * stride_cm + rn[None, :] * stride_cn)
        mask = (rm < M)[:, None] & (rn < N)[None, :]
        tl.store(C_, acc, mask=mask)


@libentry()
@triton.jit(
    do_not_specialize=[
        "iters_per_pid",
        "iters_remaining",
        "iters_per_tile",
    ],
)
def first_wave(
    A,
    B,
    C,
    M,
    N,
    K,
    locks,
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    iters_per_pid,
    iters_remaining,
    iters_per_tile,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_K: tl.constexpr,
    GROUP_M: tl.constexpr,
    EVEN_K: tl.constexpr,
):
    pid = tle.program_id(0)  # pid range from 0 to sm_count
    start_iter = pid * iters_per_pid + tl.minimum(pid, iters_remaining)
    last_iter = (pid + 1) * iters_per_pid + tl.minimum(pid + 1, iters_remaining)
    while start_iter < last_iter:
        iter_offset_in_tile = start_iter % iters_per_tile
        # Iterate over the K axis. Recalculate end_iter as M/N may change during the iteration.
        end_iter = tl.minimum(
            start_iter + (iters_per_tile - iter_offset_in_tile), last_iter
        )

        tile_id = start_iter // iters_per_tile

        pid_m, pid_n = swizzle_tile(tile_id, M, N, BLOCK_M, BLOCK_N, GROUP_M)

        rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
        rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
        rk = tl.arange(0, BLOCK_K)

        ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
        rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)

        A_base = (
            A
            + ram[:, None] * stride_am
            + rk[None, :] * stride_ak
            + BLOCK_K * stride_ak * iter_offset_in_tile
        )
        B_base = (
            B
            + rk[:, None] * stride_bk
            + rbn[None, :] * stride_bn
            + BLOCK_K * stride_bk * iter_offset_in_tile
        )

        acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)

        for current_iter in range(start_iter, end_iter):
            if EVEN_K:
                a = tl.load(A_base)
                b = tl.load(B_base)
            else:
                k_offset_in_tile = (current_iter % iters_per_tile) * BLOCK_K
                k_mask = (k_offset_in_tile + rk) < K
                a = tl.load(A_base, mask=k_mask[None, :], other=0.0)
                b = tl.load(B_base, mask=k_mask[:, None], other=0.0)

            acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)
            A_base += BLOCK_K * stride_ak
            B_base += BLOCK_K * stride_bk

        # last iteration of the tile always happens before its start on another SM
        if end_iter % iters_per_tile == 0:
            C_ptr = C + (
                rm[:, None] * stride_cm + rn[None, :] * stride_cn
            )  # compute inside the if/else to avoid spilling!
            mask = (rm < M)[:, None] & (rn < N)[None, :]
            tl.store(C_ptr, acc, mask=mask)
            if iter_offset_in_tile != 0:  # only if tile has been partially processed
                tl.atomic_xchg(locks + tile_id, 1)
        else:
            while tl.atomic_cas(locks + tile_id, 1, 1) != 1:
                pass
            C_ptr = C + (
                rm[:, None] * stride_cm + rn[None, :] * stride_cn
            )  # compute inside the if/else to avoid spilling!
            mask = (rm < M)[:, None] & (rn < N)[None, :]
            tl.atomic_add(C_ptr, acc, mask=mask, sem="relaxed")
        # next round
        start_iter = end_iter


@libentry()
@triton.jit(
    do_not_specialize=[
        "iters_per_pid",
        "iters_remaining",
        "iters_per_tile",
    ],
)
def first_wave_for_bf16(
    A,
    B,
    C,
    P,
    M,
    N,
    K,
    locks,
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    iters_per_pid,
    iters_remaining,
    iters_per_tile,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_K: tl.constexpr,
    GROUP_M: tl.constexpr,
    EVEN_K: tl.constexpr,
):
    pid = tle.program_id(0)  # pid range from 0 to sm_count
    start_iter = pid * iters_per_pid + tl.minimum(pid, iters_remaining)
    last_iter = (pid + 1) * iters_per_pid + tl.minimum(pid + 1, iters_remaining)
    while start_iter < last_iter:
        iter_offset_in_tile = start_iter % iters_per_tile
        # Iterate over the K axis. Recalculate end_iter as M/N may change during the iteration.
        end_iter = tl.minimum(
            start_iter + (iters_per_tile - iter_offset_in_tile), last_iter
        )

        tile_id = start_iter // iters_per_tile

        pid_m, pid_n = swizzle_tile(tile_id, M, N, BLOCK_M, BLOCK_N, GROUP_M)

        rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
        rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
        rk = tl.arange(0, BLOCK_K)

        ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
        rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)

        A_base = (
            A
            + ram[:, None] * stride_am
            + rk[None, :] * stride_ak
            + BLOCK_K * stride_ak * iter_offset_in_tile
        )
        B_base = (
            B
            + rk[:, None] * stride_bk
            + rbn[None, :] * stride_bn
            + BLOCK_K * stride_bk * iter_offset_in_tile
        )

        acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)

        for current_iter in range(start_iter, end_iter):
            if EVEN_K:
                a = tl.load(A_base)
                b = tl.load(B_base)
            else:
                k_offset_in_tile = (current_iter % iters_per_tile) * BLOCK_K
                k_mask = (k_offset_in_tile + rk) < K
                a = tl.load(A_base, mask=k_mask[None, :], other=0.0)
                b = tl.load(B_base, mask=k_mask[:, None], other=0.0)

            acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)
            A_base += BLOCK_K * stride_ak
            B_base += BLOCK_K * stride_bk

        rm1 = tl.arange(0, BLOCK_M)
        rn1 = tl.arange(0, BLOCK_N)

        # the first situation: not the starting parts. only need to store the data on P
        if start_iter % iters_per_tile != 0:
            P_ptr = (
                P + pid * BLOCK_M * BLOCK_N + (rm1[:, None] * BLOCK_N + rn1[None, :])
            )
            tl.store(P_ptr, acc, cache_modifier=".cg")
            # tl.debug_barrier()
            tl.atomic_xchg(locks + pid, 1)
        else:  # the first part of certain grids. shoud read datas and merge datas
            next_pid = pid + 1
            stop_loading_iter = start_iter + iters_per_tile
            end = end_iter
            while end < stop_loading_iter:
                while tl.atomic_cas(locks + next_pid, 1, 1) != 1:
                    pass
                P_ptr = (
                    P
                    + next_pid * BLOCK_M * BLOCK_N
                    + (rm1[:, None] * BLOCK_N + rn1[None, :])
                )
                acc += tl.load(P_ptr, cache_modifier=".cg")
                end += iters_per_pid + (next_pid < iters_remaining)
                next_pid += 1

            # acc = acc.to(C.dtype.element_ty)  #
            C_ = C + (rm[:, None] * stride_cm + rn[None, :] * stride_cn)
            mask = (rm < M)[:, None] & (rn < N)[None, :]
            tl.store(C_, acc, mask=mask)
        start_iter = end_iter


@libentry()
@triton.jit
def classic_mm(
    A,
    B,
    C,
    M,
    N,
    K,
    stride_am,
    stride_ak,
    stride_bk,
    stride_bn,
    stride_cm,
    stride_cn,
    total_tiles_streamk,
    BLOCK_M: tl.constexpr,
    BLOCK_N: tl.constexpr,
    BLOCK_K: tl.constexpr,
    GROUP_M: tl.constexpr,
):
    # first wave has done more tiles than there are SMs, we adjust pid
    tile_id = tle.program_id(0) + total_tiles_streamk
    pid_m, pid_n = swizzle_tile(tile_id, M, N, BLOCK_M, BLOCK_N, GROUP_M)

    # do matrix multiplication
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    # pointers
    ram = tl.max_contiguous(tl.multiple_of(rm % M, BLOCK_M), BLOCK_M)
    rbn = tl.max_contiguous(tl.multiple_of(rn % N, BLOCK_N), BLOCK_N)
    prev_multiple = prev_multiple_of(K, BLOCK_K)

    acc = tl.zeros((BLOCK_M, BLOCK_N), dtype=tl.float32)
    for start_k in range(0, prev_multiple, BLOCK_K):
        rk = start_k + tl.arange(0, BLOCK_K)
        a = tl.load(A + (ram[:, None] * stride_am + rk[None, :] * stride_ak))
        b = tl.load(B + (rk[:, None] * stride_bk + rbn[None, :] * stride_bn))
        if a.dtype != b.dtype:
            a = a.to(C.dtype.element_ty)
            b = b.to(C.dtype.element_ty)
        acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)

    # loop peeling
    rk = prev_multiple + tl.arange(0, BLOCK_K)
    mask_k = rk < K
    a = tl.load(
        A + (ram[:, None] * stride_am + rk[None, :] * stride_ak), mask=mask_k[None, :]
    )
    b = tl.load(
        B + (rk[:, None] * stride_bk + rbn[None, :] * stride_bn), mask=mask_k[:, None]
    )
    if a.dtype != b.dtype:
        a = a.to(C.dtype.element_ty)
        b = b.to(C.dtype.element_ty)
    acc += tl.dot(a, b, out_dtype=tl.float32, allow_tf32=False)

    acc = acc.to(C.dtype.element_ty)
    # rematerialize rm and rn to save registers
    rm = pid_m * BLOCK_M + tl.arange(0, BLOCK_M)
    rn = pid_n * BLOCK_N + tl.arange(0, BLOCK_N)
    C = C + (rm[:, None] * stride_cm + rn[None, :] * stride_cn)
    mask = (rm < M)[:, None] & (rn < N)[None, :]
    # handles write-back with reduction-splitting
    tl.store(C, acc, mask=mask)


def streamk_mm(a, b, c, M, N, K, sm_count=108):
    logger.debug(
        "GEMS MM, [mm scenario]: streamk, [shape info]: [-, %s, %s, %s](batch, M, N, K), "
        "[A column-major]: %s, [B column-major]: %s",
        M,
        N,
        K,
        a.stride(0) == 1,
        b.stride(0) == 1,
    )
    # TODO: change the hard code to tuning config
    BLOCK_M, BLOCK_N, BLOCK_K = 128, 128, 128
    num_stages = 3
    num_warps = 8
    GROUP_M = 8
    number_blocks_m = triton.cdiv(M, BLOCK_M)
    number_blocks_n = triton.cdiv(N, BLOCK_N)

    total_tiles = number_blocks_m * number_blocks_n
    iters_per_tile = triton.cdiv(K, BLOCK_K)
    tiles_per_wave = sm_count

    number_cooperative_tiles = total_tiles % tiles_per_wave
    number_other_tiles = total_tiles - number_cooperative_tiles
    if number_other_tiles > 0 and number_cooperative_tiles < sm_count * 0.5:
        number_cooperative_tiles = number_cooperative_tiles + tiles_per_wave
    elif number_other_tiles > 0 and number_cooperative_tiles > sm_count * 0.8:
        number_cooperative_tiles = 0

    if number_cooperative_tiles > 0:
        # mini wave
        total_iters_streamk = number_cooperative_tiles * iters_per_tile
        iters_per_pid = total_iters_streamk // tiles_per_wave
        iters_remaining = total_iters_streamk % tiles_per_wave
        even_k = K % BLOCK_K == 0

        if a.dtype == torch.bfloat16:
            locks = torch.zeros((tiles_per_wave,), device=a.device, dtype=torch.int32)
            P = torch.empty(
                (tiles_per_wave, BLOCK_M, BLOCK_N), device=a.device, dtype=torch.float32
            )
            # with torch_device_fn.device(a.device):
            first_wave_for_bf16[(tiles_per_wave,)](
                a,
                b,
                c,
                P,
                M,
                N,
                K,
                locks,
                a.stride(0),
                a.stride(1),
                b.stride(0),
                b.stride(1),
                c.stride(0),
                c.stride(1),
                iters_per_pid=iters_per_pid,
                iters_remaining=iters_remaining,
                iters_per_tile=iters_per_tile,
                BLOCK_M=BLOCK_M,
                BLOCK_N=BLOCK_N,
                BLOCK_K=BLOCK_K,
                GROUP_M=GROUP_M,
                EVEN_K=even_k,
                num_stages=num_stages,
                num_warps=num_warps,
            )
            # logger.debug(f"{k1.n_regs} registers used, {k1.n_spills} spills")
            # logger.debug(f"shared memory: {k1.metadata.shared} bytes")
        else:
            locks = torch.zeros(
                (number_cooperative_tiles,), device=a.device, dtype=torch.int32
            )
            first_wave[(tiles_per_wave,)](
                a,
                b,
                c,
                M,
                N,
                K,
                locks,
                a.stride(0),
                a.stride(1),
                b.stride(0),
                b.stride(1),
                c.stride(0),
                c.stride(1),
                iters_per_pid=iters_per_pid,
                iters_remaining=iters_remaining,
                iters_per_tile=iters_per_tile,
                BLOCK_M=BLOCK_M,
                BLOCK_N=BLOCK_N,
                BLOCK_K=BLOCK_K,
                GROUP_M=GROUP_M,
                EVEN_K=even_k,
                num_stages=num_stages,
                num_warps=num_warps,
            )
            # logger.debug(f"{k1.n_regs} registers used, {k1.n_spills} spills")
            # logger.debug(f"shared memory: {k1.metadata.shared} bytes")

    classic_mm[(total_tiles - number_cooperative_tiles,)](
        a,
        b,
        c,
        M,
        N,
        K,
        a.stride(0),
        a.stride(1),
        b.stride(0),
        b.stride(1),
        c.stride(0),
        c.stride(1),
        total_tiles_streamk=number_cooperative_tiles,
        BLOCK_M=BLOCK_M,
        BLOCK_N=BLOCK_N,
        BLOCK_K=BLOCK_K,
        GROUP_M=GROUP_M,
        num_stages=num_stages,
        num_warps=num_warps,
    )
    # logger.debug(f"{k2.n_regs} registers used, {k2.n_spills} spills")
    # logger.debug(f"shared memory: {k2.metadata.shared} bytes")
    return c



"""
Compiled Ops - 100% Autonomous Op Resolution for Compiled Mode

This module provides ALL operations for CompiledSequence without any dependency
on NativeATenDispatcher. The compiled mode is 100% independent from native mode.

Key Features:
- Pre-compiled function factories that capture all special logic at compile time
- Zero runtime isinstance() checks
- Direct PyTorch ATen calls via torch.ops.aten
- All multi-resolution fixes built-in
- All stability fixes built-in

ZERO DEPENDENCY: This module does NOT import from sequential_dispatcher.py
"""

import math
import torch
import torch.nn.functional as F
from typing import Any, Callable, Dict, Optional


def _cast_attn_mask(mask, query):
    """Cast attention mask to query dtype if needed. Shared across attention variants."""
    if mask is not None and isinstance(mask, torch.Tensor) and mask.dtype != query.dtype:
        return mask.to(query.dtype)
    return mask


def _safe_is_causal(val: Any) -> bool:
    """Convert is_causal to Python bool. Handles tensor, int, bool, and float args.

    CRITICAL: Flex FlowMatch passes is_causal as a tensor. DtypeEngine may
    downcast it to fp16. Must extract bool value or default to False.
    """
    if isinstance(val, bool):
        return val
    if isinstance(val, torch.Tensor):
        return bool(val.item()) if val.numel() == 1 else False
    if isinstance(val, (int, float)):
        return bool(val)
    return False


def _safe_dropout(val: Any) -> float:
    """Convert dropout_p to Python float. Handles tensor and numeric args."""
    if isinstance(val, (int, float)):
        return float(val)
    if isinstance(val, torch.Tensor):
        return float(val.item()) if val.numel() == 1 else 0.0
    return 0.0


def _tensor_to_int(val: Any) -> Any:
    """Convert 0-d tensor to Python int if needed."""
    if isinstance(val, torch.Tensor) and val.ndim == 0:
        return int(val.item())
    return val


def _tensor_to_int_or_none(val: Any) -> Any:
    """Convert 0-d tensor to Python int, or return None if None."""
    if val is None:
        return None
    return _tensor_to_int(val)


# ============================================================================
# OP RESOLVER - Main Entry Point
# ============================================================================

class CompiledOpResolver:
    """
    Autonomous op resolution for compiled mode.

    This is the ONLY source of truth for compiled mode operations.
    No dependency on NativeATenDispatcher.
    """

    def __init__(self, device: torch.device, dtype: torch.dtype, graph_dtype: Optional[torch.dtype] = None):
        self.device = device
        self.dtype = dtype
        self._op_cache: Dict[str, Callable] = {}

        # DtypeEngine: single entry point for all dtype decisions
        from neurobrix.core.dtype.engine import DtypeEngine
        self.dtype_engine = DtypeEngine(dtype, graph_dtype=graph_dtype)

    # ========================================================================
    # PUBLIC API
    # ========================================================================

    def get_op_func(self, op_name: str, attrs: Dict[str, Any]) -> Callable:
        """
        Get the function for an operation with all logic pre-baked.

        Args:
            op_name: The bare operation name (without aten:: prefix)
            attrs: The operation attributes from graph.json

        Returns:
            A callable that executes the operation (AMP-wrapped if applicable)
        """
        # lift_fresh is special — not dtype-related
        if op_name == "lift_fresh":
            return self._make_lift_fresh(attrs)

        # _to_copy and all other ops go through DtypeEngine.compile_op()
        if op_name == "_to_copy":
            return self.dtype_engine.compile_op(f"aten::{op_name}", None, attrs)

        # Resolve the raw function via special handlers or standard lookup
        func = self._resolve_op_func(op_name, attrs)

        # DtypeEngine wraps with AMP casting at compile time
        # Custom ops already have their prefix; only add aten:: for standard ops
        dtype_op_type = op_name if "::" in op_name else f"aten::{op_name}"
        return self.dtype_engine.compile_op(dtype_op_type, func, attrs)

    def _resolve_op_func(self, op_name: str, attrs: Dict[str, Any]) -> Callable:
        """Resolve raw op function (before DtypeEngine wrapping)."""
        if op_name == "rsqrt":
            return self._make_rsqrt()

        if op_name == "slice":
            return self._make_slice()

        if op_name == "narrow":
            return self._make_narrow()

        if op_name == "select":
            return self._make_select()

        if op_name in ("squeeze", "unsqueeze"):
            return self._make_squeeze_unsqueeze(op_name)

        if op_name == "transpose":
            return self._make_transpose()

        if op_name in ("split", "chunk"):
            return self._make_split_chunk(op_name)

        if op_name == "embedding":
            return self._make_embedding()

        if op_name in ("gather", "index_select"):
            return self._make_gather_index_select(op_name)

        if op_name in ("index_add", "scatter", "scatter_add"):
            return self._make_scatter_ops(op_name)

        # RMSNorm (pattern-reassembled)
        if op_name == "custom::rms_norm" or op_name == "rms_norm":
            return self._make_rms_norm(attrs)

        # Attention ops
        if "scaled_dot_product" in op_name and "attention" in op_name:
            return self._make_attention(op_name, attrs)

        # Upsample ops (multi-resolution fix)
        if op_name in ("upsample_nearest2d", "upsample_bilinear2d", "upsample_bicubic2d"):
            return self._make_upsample(op_name)

        # Expand (multi-resolution fix)
        if op_name == "expand":
            return self._make_expand()

        # Cat (Gemma2 scalar handling)
        if op_name == "cat":
            return self._make_cat()

        # View/Reshape (KV cache compatibility - transpose returns non-contiguous tensors)
        if op_name in ("view", "_unsafe_view"):
            return self._make_view_reshape()

        # Standard ops - resolve via PyTorch
        return self._get_standard_op(op_name)

    # ========================================================================
    # SPECIAL OP FACTORIES
    # ========================================================================

    def _make_lift_fresh(self, attrs: Dict[str, Any]) -> Callable:
        """
        Create lift_fresh function with all logic pre-compiled.

        lift_fresh is used by FakeTensorMode to mark tensors as "fresh".
        Graph capture sometimes incorrectly links to wrong tensor, causing
        shape mismatches. This handles all those cases.

        CRITICAL: Sana/Gemma2 text encoders have KV-cache patterns where:
        - input_shapes: [[1, 506, 1024]] (real tensor)
        - output_shapes: [[0]] (FakeTensor capture error)

        For these cases, we MUST passthrough the input - the [0] output shape
        is a tracer bug, not the actual expected output.
        """
        output_shapes = attrs.get("output_shapes", [])
        input_shapes = attrs.get("input_shapes", [])
        expected_output_shape = output_shapes[0] if output_shapes else None
        expected_input_shape = input_shapes[0] if input_shapes else None

        # Case 1: Graph expected scalar (shape=[])
        if expected_output_shape == []:
            def lift_fresh_scalar(inp):
                if inp is None:
                    return None
                if not isinstance(inp, torch.Tensor) or inp.ndim == 0:
                    return inp
                # Multi-dim tensor but expected scalar - compute scale
                # This handles embedding *= sqrt(hidden_size) where tracer captured
                # the scalar but runtime produces the weight tensor instead.
                # UNIVERSAL: derive hidden_size from tensor shape, not a lookup table.
                if len(inp.shape) >= 2:
                    hidden_size = inp.shape[-1]
                    scale = math.sqrt(hidden_size)
                    return torch.tensor(scale, dtype=inp.dtype, device=inp.device)
                return torch.tensor(1.0, dtype=inp.dtype, device=inp.device)
            return lift_fresh_scalar

        # Case 2: Graph expected empty tensor (shape=[0])
        # Return empty tensor so downstream cat() will filter it out.
        # This matches sequential_dispatcher.py behavior (lines 187-192).
        if expected_output_shape == [0]:
            def lift_fresh_empty(inp):
                if inp is None:
                    return None
                if not isinstance(inp, torch.Tensor):
                    return inp
                return torch.empty(0, dtype=inp.dtype, device=inp.device)
            return lift_fresh_empty

        # Case 3: Normal pass-through
        def lift_fresh_passthrough(inp):
            if inp is None:
                return None
            return inp
        return lift_fresh_passthrough

    def _make_rsqrt(self) -> Callable:
        """Create stable rsqrt function that prevents division by zero."""
        def rsqrt_stable(inp):
            if isinstance(inp, torch.Tensor):
                clamped = torch.clamp(inp, min=1e-10)
                return torch.rsqrt(clamped)
            return torch.rsqrt(inp)
        return rsqrt_stable

    def _make_slice(self) -> Callable:
        """Create slice wrapper that converts tensor args to Python ints."""
        raw_op = torch.ops.aten.slice.Tensor  # type: ignore[attr-defined]

        def slice_wrapper(input_tensor, dim, start=None, end=None, step=1):
            dim = _tensor_to_int(dim)
            start = _tensor_to_int_or_none(start)
            end = _tensor_to_int_or_none(end)
            step = _tensor_to_int(step)
            return raw_op(input_tensor, dim, start, end, step)  # type: ignore[operator]
        return slice_wrapper

    def _make_narrow(self) -> Callable:
        """Create narrow wrapper that converts tensor args to Python ints."""
        raw_op = torch.ops.aten.narrow  # type: ignore[attr-defined]

        def narrow_wrapper(input_tensor, dim, start, length):
            dim = _tensor_to_int(dim)
            start = _tensor_to_int(start)
            length = _tensor_to_int(length)
            return raw_op(input_tensor, dim, start, length)  # type: ignore[operator]
        return narrow_wrapper

    def _make_select(self) -> Callable:
        """Create select wrapper that converts tensor args to Python ints."""
        raw_op = torch.ops.aten.select.int  # type: ignore[attr-defined]

        def select_wrapper(input_tensor, dim, index):
            dim = _tensor_to_int(dim)
            index = _tensor_to_int(index)
            return raw_op(input_tensor, dim, index)  # type: ignore[operator]
        return select_wrapper

    def _make_squeeze_unsqueeze(self, op_name: str) -> Callable:
        """Create squeeze/unsqueeze wrapper that converts dim to Python int."""
        if op_name == "squeeze":
            raw_op = torch.squeeze
        else:
            raw_op = torch.unsqueeze

        def dim_wrapper(input_tensor, dim=None):
            if dim is not None:
                dim = _tensor_to_int(dim)
                return raw_op(input_tensor, dim=dim)
            if op_name == "squeeze":
                return raw_op(input_tensor, dim=None)  # type: ignore[call-arg]
            else:
                # unsqueeze requires dim parameter
                raise ValueError("unsqueeze requires dim parameter")
        return dim_wrapper

    def _make_transpose(self) -> Callable:
        """Create transpose wrapper that converts dims to Python ints."""
        raw_op = torch.transpose

        def transpose_wrapper(input_tensor, dim0, dim1):
            dim0 = _tensor_to_int(dim0)
            dim1 = _tensor_to_int(dim1)
            return raw_op(input_tensor, dim0, dim1)
        return transpose_wrapper

    def _make_split_chunk(self, op_name: str) -> Callable:
        """Create split/chunk wrapper that converts args to Python ints."""
        raw_op = torch.split if op_name == "split" else torch.chunk

        def split_wrapper(input_tensor, split_size_or_sections, dim=0):
            if isinstance(split_size_or_sections, torch.Tensor):
                split_size_or_sections = int(split_size_or_sections.item())
            dim = _tensor_to_int(dim)
            return raw_op(input_tensor, split_size_or_sections, dim)
        return split_wrapper

    def _make_embedding(self) -> Callable:
        """Create embedding wrapper that ensures int64 indices.

        ATen signature: torch.embedding(weight, indices, ...)
        """
        raw_op = torch.embedding

        def embedding_wrapper(weight, indices, *args, **kwargs):
            # Ensure indices are int64 (not float)
            if isinstance(indices, torch.Tensor) and indices.is_floating_point():
                indices = indices.long()
            return raw_op(weight, indices, *args, **kwargs)
        return embedding_wrapper

    def _make_gather_index_select(self, op_name: str) -> Callable:
        """Create gather/index_select wrapper with int conversion."""
        if op_name == "gather":
            raw_op = torch.gather
        else:
            raw_op = torch.index_select

        def index_wrapper(input_tensor, dim, index, *args, **kwargs):
            dim = _tensor_to_int(dim)
            if isinstance(index, torch.Tensor) and index.is_floating_point():
                index = index.long()
            return raw_op(input_tensor, dim, index, *args, **kwargs)
        return index_wrapper

    def _make_scatter_ops(self, op_name: str) -> Callable:
        """Create scatter/scatter_add/index_add wrapper with int conversion."""
        if op_name == "scatter":
            raw_op = torch.scatter
        elif op_name == "scatter_add":
            raw_op = torch.scatter_add
        else:  # index_add
            raw_op = torch.index_add

        def scatter_wrapper(input_tensor, dim, index, src, *args, **kwargs):
            dim = _tensor_to_int(dim)
            if isinstance(index, torch.Tensor) and index.is_floating_point():
                index = index.long()
            return raw_op(input_tensor, dim, index, src, *args, **kwargs)
        return scatter_wrapper

    def _make_rms_norm(self, attrs: Dict[str, Any]) -> Callable:
        """
        Create RMS normalization function.

        Pattern-reassembled from: pow(x,2) → mean → add(eps) → rsqrt → mul(x) → [cast] → mul(weight)

        Calling convention: CompiledSequence calls op.func(*args, **kwargs)
        where args = (x_tensor, weight_tensor) and kwargs may include epsilon.
        Epsilon is captured in closure at compile time.
        """
        import torch
        epsilon = attrs.get("epsilon", attrs.get("kwargs", {}).get("epsilon", 1e-6))

        def _rms_norm_fn(x, weight, **kwargs):
            # Compute in fp32 for numerical stability (matches decomposed pattern)
            x_fp32 = x.to(torch.float32)
            variance = x_fp32.pow(2).mean(-1, keepdim=True)
            x_normed = x_fp32 * torch.rsqrt(variance + epsilon)
            return x_normed.to(weight.dtype) * weight

        return _rms_norm_fn

    def _make_attention(self, op_name: str, attrs: Dict[str, Any]) -> Callable:
        """
        Create scaled_dot_product_attention function with variant handling.

        Handles ATen SDPA variants with different return signatures:
        - _scaled_dot_product_flash_attention_for_cpu → (output, lse)
        - _scaled_dot_product_efficient_attention     → (output, lse, seed, offset)
        - _scaled_dot_product_flash_attention         → (output, lse, seed, offset)
        - scaled_dot_product_attention                → output

        NeuroBrix executes the op as traced. DtypeEngine handles dtype.
        """
        raw_name = op_name.replace("aten::", "")
        base_name = raw_name.split(".")[0]

        if base_name == "_scaled_dot_product_flash_attention_for_cpu":
            def flash_cpu_attention(q, k, v, dropout_p=0.0, is_causal=False, **kwargs):
                scale = kwargs.get("scale", None)
                attn_mask = _cast_attn_mask(kwargs.get("attn_mask", None), q)
                output = F.scaled_dot_product_attention(
                    q, k, v,
                    attn_mask=attn_mask,
                    dropout_p=_safe_dropout(dropout_p),
                    is_causal=_safe_is_causal(is_causal),
                    scale=scale,
                )
                lse = torch.zeros((q.shape[0], q.shape[1], q.shape[2]),
                                  device=q.device, dtype=q.dtype)
                return output, lse
            return flash_cpu_attention

        if base_name in ("_scaled_dot_product_efficient_attention",
                         "_scaled_dot_product_flash_attention"):
            def efficient_attention(q, k, v, attn_bias=None, compute_lse=False,
                                   dropout_p=0.0, is_causal=False, scale=None, *args):
                q = q.contiguous()
                k = k.contiguous()
                v = v.contiguous()
                attn_bias = _cast_attn_mask(attn_bias, q)
                output = F.scaled_dot_product_attention(
                    q, k, v,
                    attn_mask=attn_bias,
                    dropout_p=_safe_dropout(dropout_p),
                    is_causal=_safe_is_causal(is_causal),
                    scale=scale,
                )
                lse = torch.zeros((q.shape[0], q.shape[1], q.shape[2]),
                                  device=q.device, dtype=q.dtype)
                philox_seed = torch.tensor(0, device=q.device, dtype=torch.int64)
                philox_offset = torch.tensor(0, device=q.device, dtype=torch.int64)
                return output, lse, philox_seed, philox_offset
            return efficient_attention

        # Standard SDPA — handles pattern-reassembled ops + native SDPA
        def standard_attention(q, k, v, *args, **kwargs):
            # K may be transposed [B,H,D,S] from pattern reassembly — fix to [B,H,S,D]
            if k.ndim == 4 and q.ndim == 4 and k.shape[-2] != q.shape[-2]:
                k = k.transpose(-2, -1)

            is_causal = _safe_is_causal(kwargs.pop('is_causal', False))
            kwargs.pop('dropout_p', None)
            scale = kwargs.pop('scale', None)

            attn_mask = None
            if args and isinstance(args[0], torch.Tensor):
                attn_mask = args[0]
                seq_q, seq_k = q.shape[2], k.shape[2]
                if attn_mask.shape[-2] > seq_q or attn_mask.shape[-1] > seq_k:
                    attn_mask = attn_mask[..., :seq_q, :seq_k].contiguous()

            if attn_mask is not None:
                attn_mask = _cast_attn_mask(attn_mask, q)

            return F.scaled_dot_product_attention(
                q, k, v, attn_mask=attn_mask,
                is_causal=is_causal, scale=scale,
            )
        return standard_attention

    def _make_upsample(self, op_name: str) -> Callable:
        """
        Create upsample function with multi-resolution fix.

        Graph contains hardcoded output_size from trace time. When scale factors
        are available, we recompute output_size from actual input dimensions.
        """
        if op_name == "upsample_nearest2d":
            raw_op = F.interpolate
            mode = "nearest"
        elif op_name == "upsample_bilinear2d":
            raw_op = F.interpolate
            mode = "bilinear"
        else:  # upsample_bicubic2d
            raw_op = F.interpolate
            mode = "bicubic"

        def upsample_wrapper(input_tensor, output_size=None, scales_h=None, scales_w=None, *args, **kwargs):
            # Multi-resolution fix: if scales are provided, recompute output_size
            if scales_h is not None and scales_w is not None:
                if isinstance(scales_h, (int, float)) and isinstance(scales_w, (int, float)):
                    h_in = input_tensor.shape[2]
                    w_in = input_tensor.shape[3]
                    output_size = (int(h_in * scales_h), int(w_in * scales_w))

            if output_size is not None:
                return raw_op(input_tensor, size=output_size, mode=mode,
                             align_corners=False if mode != "nearest" else None)
            else:
                # Fallback: use scale_factor
                return raw_op(input_tensor, scale_factor=(scales_h, scales_w), mode=mode,
                             align_corners=False if mode != "nearest" else None)
        return upsample_wrapper

    def _make_expand(self) -> Callable:
        """
        Create expand function with multi-resolution fix.

        Graph contains hardcoded expand sizes from trace time. When the tensor
        has non-singleton dimensions that don't match, use actual dimensions.
        """
        def expand_wrapper(input_tensor, size, *args, **kwargs):
            if isinstance(size, (list, tuple)) and len(size) == len(input_tensor.shape):
                new_size = list(size)
                for i, (actual, target) in enumerate(zip(input_tensor.shape, size)):
                    # expand can only broadcast from size 1
                    # If actual != 1 and actual != target, use actual
                    if actual != 1 and actual != target and target != -1:
                        new_size[i] = actual
                return input_tensor.expand(*new_size)
            return input_tensor.expand(*size)
        return expand_wrapper


    def _make_cat(self) -> Callable:
        """
        Create cat function with Gemma2 scalar tensor handling.

        Gemma2 creates scalar -inf tensors for attention masking, then tries
        to cat them. 0-dimensional tensors cannot be concatenated.
        """
        def cat_wrapper(tensors, dim=0, *args, **kwargs):
            if isinstance(tensors, (list, tuple)):
                # Filter out 0-dimensional and empty tensors
                valid_tensors = []
                for t in tensors:
                    if isinstance(t, torch.Tensor):
                        if t.ndim == 0 or t.numel() == 0:
                            continue
                    valid_tensors.append(t)

                if len(valid_tensors) == 0:
                    # All filtered - return empty tensor
                    device = tensors[0].device if tensors and isinstance(tensors[0], torch.Tensor) else 'cpu'
                    return torch.tensor([], device=device)
                elif len(valid_tensors) == 1:
                    return valid_tensors[0]
                else:
                    return torch.cat(valid_tensors, dim=dim)
            return torch.cat(tensors, dim=dim)
        return cat_wrapper

    def _make_view_reshape(self) -> Callable:
        """
        Create view function that falls back to reshape for non-contiguous tensors.

        CRITICAL FIX for KV cache: After SDPA, transpose returns non-contiguous
        tensors. The traced graph uses aten::view, but view fails on non-contiguous
        tensors. Using reshape instead handles both cases correctly.

        MoE FIX: When symbolic seq_len changes between trace and runtime, view/reshape
        may fail because numel differs from the product of the target shape dims.
        The fallback tries each position as -1 (not just the last) to infer the
        changing dimension. This handles MoE routing views where the dynamic dim
        can be in any position.
        """
        def view_or_reshape(tensor, shape, *args, **kwargs):
            # Try view first (faster, no copy)
            try:
                return tensor.view(shape)
            except RuntimeError:
                try:
                    # Fallback to reshape (handles non-contiguous)
                    return tensor.reshape(shape)
                except RuntimeError:
                    # Numel mismatch: tensor has trace-time dims, shape has runtime dims.
                    # Try each position as -1 to find the symbolic (changed) dimension.
                    if len(shape) >= 2:
                        for i in range(len(shape)):
                            trial = list(shape)
                            trial[i] = -1
                            try:
                                result = tensor.reshape(trial)
                                return result
                            except RuntimeError:
                                continue
                    raise
        return view_or_reshape

    def _get_standard_op(self, op_name: str) -> Callable:
        """
        Get standard PyTorch op function.

        Resolution priority:
        1. torch.ops.aten
        2. torch namespace
        3. torch.nn.functional
        """
        if op_name in self._op_cache:
            return self._op_cache[op_name]

        # Priority 1: torch.ops.aten
        if hasattr(torch.ops.aten, op_name):
            op = getattr(torch.ops.aten, op_name)
            self._op_cache[op_name] = op
            return op

        # Priority 2: torch namespace
        if hasattr(torch, op_name):
            op = getattr(torch, op_name)
            self._op_cache[op_name] = op
            return op

        # Priority 3: torch.nn.functional
        if hasattr(F, op_name):
            op = getattr(F, op_name)
            self._op_cache[op_name] = op
            return op

        raise AttributeError(
            f"[CompiledOpResolver] Op '{op_name}' not found in "
            f"torch.ops.aten, torch, or torch.nn.functional"
        )


# ============================================================================
# SINGLETON INSTANCE - For convenience
# ============================================================================

_resolver_instance: Optional[CompiledOpResolver] = None


def get_compiled_op_resolver(device: torch.device, dtype: torch.dtype) -> CompiledOpResolver:
    """Get or create the compiled op resolver singleton."""
    global _resolver_instance
    if _resolver_instance is None or _resolver_instance.device != device or _resolver_instance.dtype != dtype:
        _resolver_instance = CompiledOpResolver(device, dtype)
    return _resolver_instance



"""
NeuroBrix Validators - Numerical Validator.

Validates numerical stability of computations.

ZERO FALLBACK: Numerical issues raise explicit NumericalDivergence.
"""

from typing import Any, Dict, List, Optional

from .base import (
    BaseValidator, SafetyLevel, ValidationResult,
    NumericalDivergence, NaNInfDetected
)
from .config import get_config


class NumericalValidator(BaseValidator):
    """
    Validates numerical stability.

    Checks:
    - NaN detection
    - Inf detection
    - Value range for dtype
    - Gradient magnitude
    - Loss stability
    """

    @property
    def name(self) -> str:
        return "NumericalValidator"

    def __init__(self, safety_level: SafetyLevel = SafetyLevel.STANDARD):
        super().__init__(safety_level)
        self.config = get_config()
        self._check_count = 0
        self._nan_history: List[int] = []
        self._inf_history: List[int] = []

    def validate(self, graph: Any, context: Dict[str, Any]) -> ValidationResult:
        """
        Validate numerical stability of graph execution.

        Args:
            graph: Execution graph
            context: Must contain:
                - 'tensors': Dict of tensor data to check
                - 'dtype': Data type for range checking

        Returns:
            ValidationResult with numerical validation status
        """
        warnings = []
        errors = []
        metrics = {}

        tensors = context.get('tensors', {})
        dtype = context.get('dtype', 'float32')

        if not tensors:
            return ValidationResult(
                passed=True,
                validator_name=self.name,
                message="No tensors to validate",
                warnings=warnings,
            )

        total_nan = 0
        total_inf = 0
        total_elements = 0

        for name, tensor in tensors.items():
            nan_count, inf_count, elem_count = self._check_tensor(
                tensor, name, dtype, warnings, errors, metrics
            )
            total_nan += nan_count
            total_inf += inf_count
            total_elements += elem_count

        # Update history for trend detection
        self._check_count += 1
        self._nan_history.append(total_nan)
        self._inf_history.append(total_inf)

        # Keep history bounded
        if len(self._nan_history) > 100:
            self._nan_history = self._nan_history[-100:]
            self._inf_history = self._inf_history[-100:]

        # Detect trends
        if self.safety_level >= SafetyLevel.STANDARD:
            self._check_trends(warnings, metrics)

        metrics['total_nan_count'] = total_nan
        metrics['total_inf_count'] = total_inf
        metrics['total_elements_checked'] = total_elements
        metrics['check_count'] = self._check_count

        if errors:
            return ValidationResult(
                passed=False,
                validator_name=self.name,
                message="\n".join(errors),
                warnings=warnings,
                metrics=metrics,
            )

        return ValidationResult(
            passed=True,
            validator_name=self.name,
            message=f"Numerical check passed: {total_elements} elements checked",
            warnings=warnings,
            metrics=metrics,
        )

    def _check_tensor(
        self,
        tensor: Any,
        name: str,
        dtype: str,
        warnings: List[str],
        errors: List[str],
        metrics: Dict[str, Any],
    ) -> tuple:
        """Check a single tensor for numerical issues."""
        config = self.config

        # Get tensor stats (framework-agnostic)
        try:
            stats = self._get_tensor_stats(tensor)
        except Exception as e:
            warnings.append(f"Could not analyze tensor '{name}': {e}")
            return 0, 0, 0

        nan_count = stats.get('nan_count', 0)
        inf_count = stats.get('inf_count', 0)
        elem_count = stats.get('numel', 0)
        max_val = stats.get('max', 0)
        min_val = stats.get('min', 0)

        # Store per-tensor metrics
        metrics[f'{name}_nan'] = nan_count
        metrics[f'{name}_inf'] = inf_count
        metrics[f'{name}_max'] = max_val
        metrics[f'{name}_min'] = min_val

        # Check NaN threshold
        if nan_count > config.nan_threshold:
            errors.append(
                f"Tensor '{name}' contains {nan_count} NaN values "
                f"(threshold: {config.nan_threshold})"
            )

        # Check Inf threshold
        if inf_count > config.inf_threshold:
            errors.append(
                f"Tensor '{name}' contains {inf_count} Inf values "
                f"(threshold: {config.inf_threshold})"
            )

        # Check value range for dtype
        dtype_max = config.get_dtype_max(dtype)
        if abs(max_val) > dtype_max or abs(min_val) > dtype_max:
            warnings.append(
                f"Tensor '{name}' values near dtype limit: "
                f"[{min_val:.2e}, {max_val:.2e}] (max: {dtype_max:.2e})"
            )

        return nan_count, inf_count, elem_count

    def _get_tensor_stats(self, tensor: Any) -> Dict[str, Any]:
        """Get statistics from tensor (framework-agnostic)."""
        stats = {}

        # Try to get numel
        if hasattr(tensor, 'numel'):
            stats['numel'] = tensor.numel()
        elif hasattr(tensor, 'size'):
            import functools
            import operator
            stats['numel'] = functools.reduce(operator.mul, tensor.size(), 1)
        else:
            stats['numel'] = 0

        # Try to detect NaN/Inf using Triton-compatible checks
        # This assumes tensor is already on GPU and we can use Triton
        try:
            # For PyTorch tensors (during development/testing only)
            if hasattr(tensor, 'isnan'):
                stats['nan_count'] = int(tensor.isnan().sum().item())
            else:
                stats['nan_count'] = 0

            if hasattr(tensor, 'isinf'):
                stats['inf_count'] = int(tensor.isinf().sum().item())
            else:
                stats['inf_count'] = 0

            if hasattr(tensor, 'max') and hasattr(tensor, 'min'):
                # Mask out inf for max/min calculation
                finite_mask = ~(tensor.isnan() | tensor.isinf())
                if finite_mask.any():
                    finite_tensor = tensor[finite_mask]
                    stats['max'] = float(finite_tensor.max().item())
                    stats['min'] = float(finite_tensor.min().item())
                else:
                    stats['max'] = float('nan')
                    stats['min'] = float('nan')
            else:
                stats['max'] = 0
                stats['min'] = 0
        except Exception:
            stats['nan_count'] = 0
            stats['inf_count'] = 0
            stats['max'] = 0
            stats['min'] = 0

        return stats

    def _check_trends(
        self,
        warnings: List[str],
        metrics: Dict[str, Any],
    ) -> None:
        """Check for concerning numerical trends."""
        if len(self._nan_history) < 10:
            return

        # Check if NaN count is increasing
        recent_nan = self._nan_history[-10:]
        if all(recent_nan[i] <= recent_nan[i+1] for i in range(len(recent_nan)-1)):
            if recent_nan[-1] > recent_nan[0]:
                warnings.append(
                    f"NaN count increasing trend: {recent_nan[0]} -> {recent_nan[-1]}"
                )
                metrics['nan_increasing_trend'] = True

        # Check if Inf count is increasing
        recent_inf = self._inf_history[-10:]
        if all(recent_inf[i] <= recent_inf[i+1] for i in range(len(recent_inf)-1)):
            if recent_inf[-1] > recent_inf[0]:
                warnings.append(
                    f"Inf count increasing trend: {recent_inf[0]} -> {recent_inf[-1]}"
                )
                metrics['inf_increasing_trend'] = True

    def check_tensor_immediate(
        self,
        tensor: Any,
        name: str = "tensor",
        raise_on_issue: bool = True,
    ) -> Dict[str, Any]:
        """
        Immediately check a tensor for NaN/Inf.

        Args:
            tensor: Tensor to check
            name: Name for error messages
            raise_on_issue: If True, raise NaNInfDetected on issues

        Returns:
            Dict with check results

        Raises:
            NaNInfDetected if issues found and raise_on_issue=True
        """
        stats = self._get_tensor_stats(tensor)

        if raise_on_issue:
            if stats.get('nan_count', 0) > 0:
                raise NaNInfDetected(
                    f"NaN detected in {name}",
                    nan_count=stats['nan_count'],
                    suggestion="Check for division by zero, log of negative, or exploding gradients",
                )
            if stats.get('inf_count', 0) > 0:
                raise NaNInfDetected(
                    f"Inf detected in {name}",
                    inf_count=stats['inf_count'],
                    suggestion="Check for overflow or very large values before operations",
                )

        return stats

    def reset_history(self) -> None:
        """Reset numerical tracking history."""
        self._check_count = 0
        self._nan_history = []
        self._inf_history = []

    def should_run(self, safety_level: SafetyLevel) -> bool:
        """Numerical validation runs at STANDARD and PARANOID levels."""
        return safety_level >= SafetyLevel.STANDARD



import builtins
import logging

import torch
import triton
import triton.language as tl

# from flag_gems import runtime
from flag_gems.runtime import torch_device_fn
from flag_gems.utils import dim_compress, libentry
from flag_gems.utils import triton_lang_extension as tle

from ..utils.block_size_utils import get_block_size_1d

logger = logging.getLogger("flag_gems").getChild(__name__.lstrip("."))


@libentry()
@triton.jit
def mean_kernel_1(
    inp,
    mid,
    M,
    BLOCK_SIZE: tl.constexpr,
):
    pid = tle.program_id(0)
    offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
    inp_ptrs = inp + offset
    mask = offset < M
    inp_val = tl.load(inp_ptrs, mask=mask, other=0.0)
    sum_val = tl.sum(inp_val, axis=0)
    mid_ptr = mid + pid
    tl.store(mid_ptr, sum_val)


@libentry()
@triton.jit
def mean_kernel_2(mid, out, M, MID_SIZE, BLOCK_MID: tl.constexpr):
    offset = tl.arange(0, BLOCK_MID)
    mid_ptrs = mid + offset
    mask = offset < MID_SIZE
    mid_val = tl.load(mid_ptrs, mask=mask, other=0.0)
    sum_val = tl.sum(mid_val, axis=0) / M
    tl.store(out, sum_val)


def mean(inp, *, dtype=None):
    logger.debug("GEMS MEAN")
    M = inp.numel()
    if dtype is None:
        dtype = inp.dtype
    # block_size = triton.next_power_of_2(math.ceil(math.sqrt(M)))
    block_size = get_block_size_1d(M, inp.element_size())
    mid_size = triton.cdiv(M, block_size)
    block_mid = triton.next_power_of_2(mid_size)

    mid = torch.empty((mid_size,), dtype=dtype, device=inp.device)
    out = torch.empty([], dtype=dtype, device=inp.device)

    with torch_device_fn.device(inp.device):
        mean_kernel_1[(mid_size, 1, 1)](inp, mid, M, block_size, buffer_size_limit=2048)
        if mid_size == 1:
            return (mid / M).reshape([])
        mean_kernel_2[(1, 1, 1)](
            mid, out, M, mid_size, block_mid, buffer_size_limit=2048
        )
    return out


def heur_m_block_size(args):
    return triton.next_power_of_2(triton.cdiv(args["M"], 12))  # cluster_num


def heur_n_block_size(args):
    return builtins.min(args["N"], 8192)


@libentry()
# @triton.autotune(
#     configs=runtime.get_tuned_config("mean"),
#     key=["M", "N"],
# )
@triton.heuristics(
    values={
        "BLOCK_M": heur_m_block_size,
        "BLOCK_N": heur_n_block_size,
    },
)
@triton.jit
def mean_dim_kernel(X, Mean, M, N, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr):
    # Map the program id to the row of X it should compute.
    pid = tle.program_id(0) * BLOCK_M + tl.arange(0, BLOCK_M)[:, None]
    X = X + pid * N
    Mean = Mean + pid
    row_mask = pid < M

    # Compute mean
    _mean = tl.zeros([BLOCK_M, BLOCK_N], dtype=tl.float32)
    for off in range(0, N, BLOCK_N):
        cols = off + tl.arange(0, BLOCK_N)[None, :]
        col_mask = cols < N
        mask = row_mask and col_mask

        a = tl.load(X + cols, mask, other=0.0).to(tl.float32)
        _mean += a
    mean = tl.sum(_mean, axis=1) / N
    mean = mean[:, None]
    tl.store(Mean, mean, row_mask)


def mean_dim(x, dim, keepdim=False, *, dtype=None):
    logger.debug("GEMS MEAN DIM")

    if dtype is None:
        dtype = x.dtype
    if dim is None:
        out = mean(x, dtype=dtype)
        if not keepdim:
            out = out.reshape([1] * x.ndim)
        return out

    shape = list(x.shape)
    dim = [d % x.ndim for d in dim]
    x = dim_compress(x, dim)
    N = 1
    for i in dim:
        N *= shape[i]
        shape[i] = 1
    M = x.numel() // N
    out = torch.empty(shape, dtype=dtype, device=x.device)
    grid = lambda META: (triton.cdiv(M, META["BLOCK_M"]),)

    with torch_device_fn.device(x.device):
        mean_dim_kernel[grid](x, out, M, N, buffer_size_limit=2048)
    if not keepdim:
        out = out.squeeze(dim)
    return out
